@@ -23,11 +23,12 @@ from absl import logging
 from etils import epath
 import jax
 from orbax.checkpoint import checkpoint_args
-from orbax.checkpoint import future as future_lib
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import utils
 from orbax.checkpoint._src import asyncio_utils
 from orbax.checkpoint._src.checkpointers import checkpointer
+from orbax.checkpoint._src.futures import future
+from orbax.checkpoint._src.futures import synchronization
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
 from orbax.checkpoint._src.metadata import checkpoint
 from orbax.checkpoint._src.multihost import multihost
@@ -36,7 +37,11 @@ from orbax.checkpoint._src.path import atomicity
 from orbax.checkpoint._src.path import atomicity_types
 
 
+
 BarrierSyncFn = multihost.BarrierSyncFn
+_DIRECTORY_CREATION_SIGNALS = [
+    synchronization.HandlerAwaitableSignal.STEP_DIRECTORY_CREATION
+]
 
 
 def _on_commit_callback(
@@ -45,11 +50,13 @@ def _on_commit_callback(
 ):
   """Finalize atomic save and record checkpoint save metrics."""
   atomicity.on_commit_callback(
-      tmpdir, checkpoint_start_time=checkpoint_start_time
+      tmpdir,
+      checkpoint_start_time=checkpoint_start_time,
   )
+  total_duration_secs = time.time() - checkpoint_start_time
   jax.monitoring.record_event_duration_secs(
       '/jax/checkpoint/write/async/total_duration_secs',
-      time.time() - checkpoint_start_time,
+      total_duration_secs,
   )
 
 
@@ -106,7 +113,7 @@ class _AsyncManager:
   def _thread_func(
       self,
       directory: epath.Path,
-      commit_futures: Sequence[future_lib.Future],
+      commit_futures: Sequence[future.Future],
       on_commit_callback: Callable[[], None],
   ):
     """Awaits on commit futures and finalizes the checkpoint."""
@@ -122,8 +129,8 @@ class _AsyncManager:
       thread_start_time = time.time()
 
       # Wait for commit operations to complete.
-      for future in commit_futures:
-        future.result()
+      for commit_future in commit_futures:
+        commit_future.result()
       logging.info(
           '[process=%s][thread=%s] %d Handler Commit operations completed.',
           current_process,
@@ -138,10 +145,12 @@ class _AsyncManager:
       )
 
       # Log the per process storage commit latency excluding the barrier time.
+      commit_duration_secs = time.time() - thread_start_time
       jax.monitoring.record_event_duration_secs(
           '/jax/checkpoint/write/async/commit_duration_sec',
-          time.time() - thread_start_time,
+          commit_duration_secs,
       )
+      logging.vlog(1, 'Async Commit duration: %s seconds', commit_duration_secs)
 
       if process_count > 1:
         # All processes will wait at the barrier. When all processes are at the
@@ -172,10 +181,12 @@ class _AsyncManager:
             )
         )
 
+      thread_duration_secs = time.time() - thread_start_time
       jax.monitoring.record_event_duration_secs(
           '/jax/checkpoint/write/async/thread_duration_sec',
-          time.time() - thread_start_time,
+          thread_duration_secs,
       )
+      logging.vlog(1, 'Async thread duration: %s seconds', thread_duration_secs)
       logging.info(
           '[process=%s][thread=%s] Background save thread done.',
           current_process,
@@ -194,7 +205,7 @@ class _AsyncManager:
   def start_async_commit(
       self,
       directory: epath.Path,
-      commit_futures: Sequence[future_lib.Future],
+      commit_futures: Sequence[future.Future],
       on_commit_callback: Callable[[], None],
   ):
     """Completes checkpoint save in a background thread."""
@@ -296,6 +307,9 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
     self._primary_host = multiprocessing_options.primary_host
     self._active_processes = multiprocessing_options.active_processes
     self._post_finalization_callback = async_options.post_finalization_callback
+    self._create_directories_asynchronously = (
+        async_options.create_directories_asynchronously
+    )
     barrier_sync_key_prefix = (
         ''
         if multiprocessing_options.barrier_sync_key_prefix is None
@@ -304,9 +318,8 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
     self._barrier_sync_key_prefix = barrier_sync_key_prefix
     self._file_options = file_options
     self._metadata_store = (
-        checkpoint_metadata_store or checkpoint.metadata_store(
-            enable_write=True
-        )
+        checkpoint_metadata_store
+        or checkpoint.metadata_store(enable_write=True)
     )
     self._temporary_path_class = temporary_path_class
     timeout_secs = timeout_secs or async_options.timeout_secs
@@ -323,13 +336,20 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
         primary_host=multiprocessing_options.primary_host,
         barrier_sync_key_prefix=barrier_sync_key_prefix,
     )
+    self._multiprocessing_options = multiprocessing_options
 
   async def _save(
-      self, directory: epath.PathLike, *args, force: bool = False, **kwargs
+      self,
+      directory: epath.PathLike,
+      *args,
+      force: bool = False,
+      custom_metadata: dict[str, Any] | None = None,
+      **kwargs
   ):
     checkpoint_start_time = time.time()
     directory = epath.Path(directory)
     self.wait_until_finished()
+    self.synchronize_next_awaitable_signal_operation_id()
 
     jax.monitoring.record_event('/jax/orbax/write/async/start')
     logging.info(
@@ -351,19 +371,37 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
       else:
         raise ValueError(f'Destination {directory} already exists.')
 
-    tmpdir = await self.create_temporary_path(directory)
+    commit_ops = []
+    tmpdir = self.get_temporary_path(directory)
+    if self._create_directories_asynchronously:
+      commit_ops.append(
+          atomicity.create_all_async(
+              [tmpdir],
+              completion_signals=_DIRECTORY_CREATION_SIGNALS,
+              multiprocessing_options=self._multiprocessing_options,
+          )
+      )
+    else:
+      await self.create_temporary_path(tmpdir)
     # Run copy ops.
     # Try to save using new CheckpointArgs API if supported by the handler.
     ckpt_args = checkpointer.construct_checkpoint_args(
         self._handler, True, *args, **kwargs
     )
-    commit_ops = await self._handler.async_save(tmpdir.get(), args=ckpt_args)
+    commit_ops.extend(
+        await self._handler.async_save(tmpdir.get(), args=ckpt_args) or []
+    )
     commit_ops, _ = jax.tree.flatten(commit_ops)
     commit_ops = [op for op in commit_ops if op is not None]
 
     # Directory is the final directory.
     def _callback() -> None:
-      logging.info(
+      if utils.is_primary_host(self._primary_host):
+        # Update StepMetadata after the handler save is complete.
+        # (blocking write)
+        self._save_step_metadata(tmpdir.get(), custom_metadata=custom_metadata)
+      logging.vlog(
+          1,
           '[process=%s][thread=%s] Async Save Callback [1/3]: Finalizing'
           ' Handler: %s on %s',
           multihost.process_index(),
@@ -371,8 +409,10 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
           self._handler,
           tmpdir.get(),
       )
+      # Finalize does a final StepMetadata update.
       self._handler.finalize(tmpdir.get())
-      logging.info(
+      logging.vlog(
+          1,
           '[process=%s][thread=%s] Async Save Callback [2/3]: Running'
           ' post_finalization_callback: %s on %s',
           multihost.process_index(),
@@ -382,28 +422,49 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
       )
       if self._post_finalization_callback is not None:
         self._post_finalization_callback()
-      logging.info(
+      logging.vlog(
+          1,
           '[process=%s][thread=%s] Async Save Callback [3/3]: Finalizing'
           ' checkpoint directory: %s',
           multihost.process_index(),
           threading.current_thread().name,
           tmpdir.get(),
       )
-      _on_commit_callback(tmpdir, checkpoint_start_time)
+      _on_commit_callback(
+          tmpdir,
+          checkpoint_start_time,
+      )
+      logging.info(
+          'Finished asynchronous save (blocking + background) in %.2f seconds'
+          ' to %s',
+          time.time() - checkpoint_start_time,
+          directory,
+      )
 
     self._async_manager.start_async_commit(
         directory,
         commit_futures=commit_ops,
         on_commit_callback=_callback,
     )
-
+    blocking_duration_secs = time.time() - checkpoint_start_time
     jax.monitoring.record_event_duration_secs(
         '/jax/checkpoint/write/async/blocking_duration_secs',
-        time.time() - checkpoint_start_time,
+        blocking_duration_secs,
+    )
+    logging.info(
+        'Finished blocking save in %.2f seconds. Continuing to save'
+        ' asynchronously to %s.',
+        blocking_duration_secs,
+        directory,
     )
 
   def save(
-      self, directory: epath.PathLike, *args, force: bool = False, **kwargs
+      self,
+      directory: epath.PathLike,
+      *args,
+      force: bool = False,
+      custom_metadata: dict[str, Any] | None = None,
+      **kwargs
   ):
     """Saves the given item to the provided directory.
 
@@ -419,13 +480,23 @@ class AsyncCheckpointer(checkpointer.Checkpointer):
       *args: additional args to provide to the CheckpointHandler's save method.
       force: if True, allows overwriting an existing directory. May add overhead
         due to the need to delete any existing files.
+      custom_metadata: a dictionary of custom metadata to be written to the
+        checkpoint directory via StepMetadata.
       **kwargs: additional keyword args to provide to the CheckpointHandler's
         save method.
 
     Raises:
       ValueError if the provided directory already exists.
     """
-    asyncio_utils.run_sync(self._save(directory, *args, force=force, **kwargs))
+    asyncio_utils.run_sync(
+        self._save(
+            directory,
+            *args,
+            force=force,
+            custom_metadata=custom_metadata,
+            **kwargs
+        )
+    )
 
   def restore(self, directory: epath.PathLike, *args, **kwargs) -> Any:
     """See superclass documentation."""
