@@ -22,7 +22,6 @@ customized, and is delegated to the `TypeHandler` class.
 from __future__ import annotations
 
 import asyncio
-from concurrent import futures
 import dataclasses
 import functools
 import json
@@ -36,11 +35,13 @@ from etils import epath
 import humanize
 import jax
 from orbax.checkpoint import checkpoint_args
-from orbax.checkpoint import future
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import utils
 from orbax.checkpoint._src import asyncio_utils
+from orbax.checkpoint._src.futures import future
+from orbax.checkpoint._src.futures import synchronization
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
+from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
 from orbax.checkpoint._src.metadata import empty_values
 from orbax.checkpoint._src.metadata import tree as tree_metadata
 from orbax.checkpoint._src.multihost import multihost
@@ -49,6 +50,7 @@ from orbax.checkpoint._src.serialization import serialization
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
 from orbax.checkpoint._src.serialization import type_handlers
 from orbax.checkpoint._src.serialization import types
+from orbax.checkpoint._src.tree import types as tree_types
 from orbax.checkpoint._src.tree import utils as tree_utils
 import tensorstore as ts
 
@@ -282,6 +284,9 @@ class BasePyTreeCheckpointHandler(
       pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
           tree_metadata.PYTREE_METADATA_OPTIONS
       ),
+      array_metadata_validator: array_metadata_store_lib.Validator = (
+          array_metadata_store_lib.Validator()
+      ),
   ):
     """Creates BasePyTreeCheckpointHandler.
 
@@ -301,6 +306,7 @@ class BasePyTreeCheckpointHandler(
       enable_post_merge_validation: If True, enables validation of the
         parameters after the finalize step.
       pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
+      array_metadata_validator: Validator for ArrayMetadata.
     """
     self._save_concurrent_bytes = save_concurrent_bytes
     self._restore_concurrent_bytes = restore_concurrent_bytes
@@ -310,18 +316,24 @@ class BasePyTreeCheckpointHandler(
     self._type_handler_registry = type_handler_registry
     self._enable_post_merge_validation = enable_post_merge_validation
     self._pytree_metadata_options = pytree_metadata_options
+    # Get ArrayMetadata Store from TypeHandler for jax.Array.
+    # ArrayMetadata persistence is only supported for jax.Array.
+    self._array_metadata_store = (
+        array_metadata_store_lib.resolve_array_metadata_store(
+            type_handler_registry
+        )
+    )
+    self._array_metadata_validator = array_metadata_validator
 
 
     jax.monitoring.record_event(
         '/jax/orbax/pytree_checkpoint_handler/init/ocdbt'
     )
-
-    self._thread_pool = futures.ThreadPoolExecutor(
-        max_workers=2, thread_name_prefix='base_pytree_ch'
-    )
     logging.info(
-        'Created BasePyTreeCheckpointHandler: pytree_metadata_options=%s',
+        'Created BasePyTreeCheckpointHandler: pytree_metadata_options=%s,'
+        ' array_metadata_store=%s',
         self._pytree_metadata_options,
+        self._array_metadata_store,
     )
 
   def get_param_names(self, item: PyTree) -> PyTree:
@@ -336,7 +348,7 @@ class BasePyTreeCheckpointHandler(
       use_ocdbt: bool = True,
       use_zarr3: Optional[bool] = None,
       ocdbt_target_data_file_size: Optional[int] = None,
-      enable_pinned_host_transfer: bool = True,
+      enable_pinned_host_transfer: bool = False,
       byte_limiter: Optional[serialization.ByteLimiter] = None,
       raise_array_data_missing_error: bool = True,
   ) -> PyTree:
@@ -436,6 +448,7 @@ class BasePyTreeCheckpointHandler(
       raise ValueError('Found empty item.')
     save_args = args.save_args
     ocdbt_target_data_file_size = args.ocdbt_target_data_file_size
+    custom_metadata = args.custom_metadata
 
     save_args = _fill_missing_save_or_restore_args(item, save_args, mode='save')
     byte_limiter = serialization.get_byte_limiter(self._save_concurrent_bytes)
@@ -451,7 +464,7 @@ class BasePyTreeCheckpointHandler(
         leaf.parent_dir == directory for leaf in jax.tree.leaves(param_infos)
     )
 
-    serialize_ops = []
+    serialize_ops = []  # List of (coros -> List of futures)
     batch_requests = batched_serialization_requests(
         item,
         param_infos,
@@ -465,20 +478,36 @@ class BasePyTreeCheckpointHandler(
       ]
       write_size, _ = _get_batch_memory_size(request.handler, request.values)
       tree_memory_size += write_size
-    # Await copy futures. Returns list of lists.
+    # Await copy futures. Returns List[List[future.Future]].
     commit_futures = await asyncio.gather(*serialize_ops)
+    # Flatten to List[future.Future].
     commit_futures, _ = jax.tree.flatten(commit_futures)
 
     if logging.vlog_is_on(1):
       logging.vlog(1, 'param_info: %s', param_infos)
       logging.vlog(1, 'save_args: %s', save_args)
 
+    save_futures = []
     if multihost.is_primary_host(self._primary_host):
-      commit_futures.append(
-          self._write_metadata_file(
-              directory, param_infos, save_args, self._use_zarr3
+      operation_id = (
+          synchronization.HandlerAwaitableSignalOperationIdGenerator.get_current_operation_id()
+      )
+      save_futures.append(
+          future.CommitFuture(
+              self._write_metadata_after_commits(
+                  commit_futures,
+                  checkpoint_dir=directory,
+                  param_infos=param_infos,
+                  save_args=save_args,
+                  custom_metadata=custom_metadata,
+                  use_zarr3=self._use_zarr3,
+                  operation_id=operation_id,
+              ),
+              name='write_metadata_after_commits',
           )
       )
+    else:
+      save_futures += commit_futures
 
     _log_io_metrics(
         tree_memory_size,
@@ -487,7 +516,7 @@ class BasePyTreeCheckpointHandler(
     )
     return [
         future.ChainedFuture(
-            commit_futures,
+            save_futures,
             functools.partial(
                 _log_io_metrics,
                 tree_memory_size,
@@ -725,14 +754,70 @@ class BasePyTreeCheckpointHandler(
     )
     return restored_item
 
+  async def _get_param_infos_with_write_shape(
+      self,
+      param_infos: PyTree,
+      checkpoint_dir: epath.Path,
+      array_metadata_store: array_metadata_store_lib.Store,
+  ) -> PyTree:
+    """Returns `param_infos` updated with `write_shape`.
+
+    Args:
+      param_infos: A PyTree of ParamInfo to be updated.
+      checkpoint_dir: The checkpoint directory where write_shape metadata is
+        saved in ArrayMetadata store.
+      array_metadata_store: The ArrayMetadata store to read write_shape metadata
+        from.
+    """
+    if not utils.is_primary_host(self._primary_host):
+      return param_infos
+    # Extract write_shape from ArrayMetadata for current process_index.
+    process_index = multihost.process_index()
+    array_metadatas = await array_metadata_store.read(
+        checkpoint_dir, process_index=process_index
+    )
+    if array_metadatas is None:
+      jax_array_param_info = type_handlers.any_jax_array_param_info(param_infos)
+      if jax_array_param_info is not None:
+        raise ValueError(
+            f'No ArrayMetadata found for process_index={process_index} in the'
+            f' checkpoint directory: {checkpoint_dir}. But input PyTree'
+            ' contains at least one jax.Array param_info:'
+            f' {jax_array_param_info}.'
+        )
+      return param_infos
+
+    assert isinstance(array_metadatas, list)
+    array_metadatas_cache = {
+        array_metadata.param_name: array_metadata
+        for array_metadata in array_metadatas
+    }
+
+    def update_param_info(param_info: types.ParamInfo) -> types.ParamInfo:
+      if not type_handlers.represents_jax_array(param_info):
+        return param_info
+      if param_info.name not in array_metadatas_cache:
+        raise ValueError(
+            f'No ArrayMetadata found for param_info: {param_info}, checkpoint'
+            f' directory: {checkpoint_dir}, process_index={process_index}.'
+        )
+      return dataclasses.replace(
+          param_info,
+          write_shape=array_metadatas_cache[param_info.name].write_shape,
+      )
+
+    return jax.tree.map(update_param_info, param_infos)
+
   def _write_metadata_file(
       self,
       directory: epath.Path,
+      *,
       param_infos: PyTree,
       save_args: PyTree,
+      custom_metadata: tree_types.JsonType | None = None,
       use_zarr3: bool = False,
   ) -> future.Future:
-    def _save_fn():
+    async def _save_fn(param_infos):
       if utils.is_primary_host(self._primary_host):
         metadata_write_start_time = time.time()
         path = directory / PYTREE_METADATA_FILE
@@ -740,6 +825,7 @@ class BasePyTreeCheckpointHandler(
             param_infos,
             save_args=save_args,
             use_zarr3=use_zarr3,
+            custom_metadata=custom_metadata,
             pytree_metadata_options=self._pytree_metadata_options,
         )
         logging.vlog(
@@ -753,9 +839,54 @@ class BasePyTreeCheckpointHandler(
             '/jax/checkpoint/write/async/metadata_write_duration_secs',
             time.time() - metadata_write_start_time,
         )
-      return 0
 
-    return self._thread_pool.submit(_save_fn)
+    return future.CommitFuture(
+        _save_fn(param_infos),
+    )
+
+  async def _write_metadata_after_commits(
+      self,
+      commit_futures: List[future.Future],
+      checkpoint_dir: epath.Path,
+      *,
+      param_infos: PyTree,
+      save_args: PyTree,
+      custom_metadata: tree_types.JsonType | None = None,
+      use_zarr3: bool,
+      operation_id: str | None = None,
+  ) -> None:
+    if not utils.is_primary_host(self._primary_host):
+      return
+    for commit_future in commit_futures:
+      await asyncio.to_thread(commit_future.result)
+    # `write_shape` is extracted from ArrayMetadata store saved during
+    # materialization of commit_futures. Then it is written to the pytree
+    # metadata.
+    # TODO(b/390465017): Simplify all metadata related code in this module after
+    # removing overriding of self._write_metadata_file() in subclasses. All
+    # metadata related code can be moved to a separate class and
+    # BasePyTreeCheckpointHandler should delegate all metadata related code to
+    # that class.
+    if self._array_metadata_store is not None:
+      param_infos = await self._get_param_infos_with_write_shape(
+          param_infos, checkpoint_dir, self._array_metadata_store
+      )
+
+    async def metadata_coro():
+      self._write_metadata_file(
+          checkpoint_dir,
+          param_infos=param_infos,
+          save_args=save_args,
+          custom_metadata=custom_metadata,
+          use_zarr3=use_zarr3,
+      ).result()
+
+    metadata_write_future = future.CommitFutureAwaitingContractedSignals(
+        metadata_coro(),
+        name='write_metadata_file',
+        operation_id=operation_id,
+    )
+    await asyncio.to_thread(metadata_write_future.result)
 
   def _read_metadata_file(
       self, directory: epath.Path
@@ -788,7 +919,8 @@ class BasePyTreeCheckpointHandler(
         pytree_metadata_options=self._pytree_metadata_options,
     )
 
-  def metadata(self, directory: epath.Path) -> Optional[PyTree]:
+
+  def metadata(self, directory: epath.Path) -> tree_metadata.TreeMetadata:
     """Returns tree metadata.
 
     The result will be a PyTree matching the structure of the saved checkpoint.
@@ -815,8 +947,14 @@ class BasePyTreeCheckpointHandler(
       tree containing metadata.
     """
     is_ocdbt_checkpoint = type_handlers.is_ocdbt_checkpoint(directory)
-    return self._read_metadata_file(directory).as_user_metadata(
-        directory, self._type_handler_registry, use_ocdbt=is_ocdbt_checkpoint
+    internal_tree_metadata = self._read_metadata_file(directory)
+    return tree_metadata.build_default_tree_metadata(
+        internal_tree_metadata.as_custom_metadata(
+            directory,
+            self._type_handler_registry,
+            use_ocdbt=is_ocdbt_checkpoint,
+        ),
+        custom_metadata=internal_tree_metadata.custom_metadata,
     )
 
   def finalize(self, directory: epath.Path) -> None:
@@ -829,24 +967,45 @@ class BasePyTreeCheckpointHandler(
     Args:
       directory: Path where the checkpoint is located.
     """
-    merge_start_time = time.time()
-    ts_context = ts_utils.get_ts_context(use_ocdbt=True)
-    asyncio_utils.run_sync(
-        type_handlers.merge_ocdbt_per_process_files(
-            directory,
-            ts_context=ts_context,
-            use_zarr3=self._use_zarr3,
-            enable_validation=self._enable_post_merge_validation,
+    finalize_coros = []
+    if self._array_metadata_store is not None:
+      if self._primary_host is None:
+        logging.log_first_n(
+            logging.WARNING,
+            '[process=%s] Skipped cross-host ArrayMetadata validation'
+            ' because all hosts are primary (e.g. local storage).',
+            1,  # log only once
+            multihost.process_index(),
+        )
+      elif utils.is_primary_host(self._primary_host):
+        finalize_coros.append(
+            array_metadata_store_lib.validate_all_array_metadatas(
+                self._array_metadata_validator,
+                self._array_metadata_store,
+                directory,
+            )
         )
-    )
-    jax.monitoring.record_event_duration_secs(
-        '/jax/checkpoint/write/async/ocdbt_merge_duration_secs',
-        time.time() - merge_start_time,
-    )
 
-  def close(self):
-    """Closes the handler. Called automatically by Checkpointer."""
-    self._thread_pool.shutdown()
+    async def merge_ocdbt_per_process_files():
+      merge_start_time = time.time()
+      ts_context = ts_utils.get_ts_context(use_ocdbt=True)
+      await type_handlers.merge_ocdbt_per_process_files(
+          directory,
+          ts_context=ts_context,
+          use_zarr3=self._use_zarr3,
+          enable_validation=self._enable_post_merge_validation,
+      )
+      jax.monitoring.record_event_duration_secs(
+          '/jax/checkpoint/write/async/ocdbt_merge_duration_secs',
+          time.time() - merge_start_time,
+      )
+
+    finalize_coros.append(merge_ocdbt_per_process_files())
+
+    async def _fn():
+      await asyncio.gather(*finalize_coros)
+
+    asyncio_utils.run_sync(_fn())
 
 
 @register_with_handler(BasePyTreeCheckpointHandler, for_save=True)
@@ -865,15 +1024,19 @@ class BasePyTreeSaveArgs(CheckpointArgs):
       indicates no maximum file size limit.  For best results, ensure
       chunk_byte_size is smaller than this value.  For more details, refer to
       https://google.github.io/tensorstore/kvstore/ocdbt/index.html#json-kvstore/ocdbt.target_data_file_size
-    enable_pinned_host_transfer: True by default. If False, disables transfer to
-      pinned host when copying from device to host, regardless of the presence
-      of pinned host memory.
+    enable_pinned_host_transfer: If False, disables transfer to pinned host when
+      copying from device to host, regardless of the presence of pinned host
+      memory.
+    custom_metadata: User-provided custom metadata. An arbitrary
+      JSON-serializable dictionary the user can use to store additional
+      information. The field is treated as opaque by Orbax.
   """
 
   item: PyTree
   save_args: Optional[PyTree] = None
   ocdbt_target_data_file_size: Optional[int] = None
-  enable_pinned_host_transfer: bool = True
+  enable_pinned_host_transfer: bool = False
+  custom_metadata: tree_types.JsonType | None = None
 
 
 @register_with_handler(BasePyTreeCheckpointHandler, for_restore=True)
