@@ -55,14 +55,18 @@ from absl import logging
 from etils import epath
 import jax
 from orbax.checkpoint import checkpoint_args
-from orbax.checkpoint import future
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint._src import asyncio_utils
 from orbax.checkpoint._src import composite
+from orbax.checkpoint._src.futures import future
+from orbax.checkpoint._src.futures import synchronization
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
 from orbax.checkpoint._src.handlers import checkpoint_handler
 from orbax.checkpoint._src.handlers import handler_registration
+from orbax.checkpoint._src.handlers import handler_type_registry
 from orbax.checkpoint._src.handlers import proto_checkpoint_handler
+from orbax.checkpoint._src.metadata import checkpoint
+from orbax.checkpoint._src.metadata import step_metadata_serialization
 from orbax.checkpoint._src.path import atomicity
 from orbax.checkpoint._src.path import atomicity_defaults
 from orbax.checkpoint._src.path import atomicity_types
@@ -71,6 +75,8 @@ CheckpointArgs = checkpoint_args.CheckpointArgs
 Future = future.Future
 CheckpointArgs = checkpoint_args.CheckpointArgs
 CheckpointHandler = checkpoint_handler.CheckpointHandler
+StepMetadata = checkpoint.StepMetadata
+CompositeItemMetadata = checkpoint.CompositeItemMetadata
 AsyncCheckpointHandler = async_checkpoint_handler.AsyncCheckpointHandler
 register_with_handler = checkpoint_args.register_with_handler
 ProtoCheckpointHandler = proto_checkpoint_handler.ProtoCheckpointHandler
@@ -78,11 +84,14 @@ ProtoSaveArgs = proto_checkpoint_handler.ProtoSaveArgs
 ProtoRestoreArgs = proto_checkpoint_handler.ProtoRestoreArgs
 AsyncSaveCoroutine = Coroutine[Any, Any, Optional[List[Future]]]
 Composite = composite.Composite
+HandlerAwaitableSignal = synchronization.HandlerAwaitableSignal
 
 
 _CONCURRENT_WORKERS = 3
-RESERVED_ITEM_NAMES = []
 
+DESCRIPTOR_ITEM_NAME = 'descriptor'
+RESERVED_ITEM_NAMES = [DESCRIPTOR_ITEM_NAME]
+_DIRECTORY_CREATION_SIGNALS = [HandlerAwaitableSignal.ITEM_DIRECTORY_CREATION]
 
 
 # TODO(b/295899152) Clean up when users are all registering `CheckpointArgs`.
@@ -192,6 +201,7 @@ class CompositeOptions:
   )
   temporary_path_class: Optional[Type[atomicity_types.TemporaryPath]] = None
   file_options: Optional[options_lib.FileOptions] = None
+  async_options: Optional[options_lib.AsyncOptions] = None
 
 
 def _get_unique_registered_items_and_handlers(
@@ -269,6 +279,7 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
   the items and handlers. Please use `handler_registry` instead.
 
   Usage::
+
     # The simplest use-case, with no handler registry provided on construction.
     checkpointer = ocp.Checkpointer(
         ocp.CompositeCheckpointHandler()
@@ -294,7 +305,7 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
     checkpointer.save(directory,
       ocp.args.Composite(
         # Will raise `ValueError: Item "state" and args "JsonSave [...]" does
-        not match with any registered handler! ...`.
+        # not match with any registered handler! ...`.
         state=ocp.args.JsonSave(pytree),
         )
     )
@@ -490,19 +501,16 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
 
     self._temporary_path_class = composite_options.temporary_path_class
     self._file_options = composite_options.file_options
+    self._async_options = (
+        composite_options.async_options or options_lib.AsyncOptions()
+    )
 
     logging.info(
         'Initialized registry %s.',
         self._handler_registry,
     )
 
-  # TODO: b/359524229 - Remove this property as it has been deprecated.
-  @property
-  def _known_handlers(self) -> Dict[str, Optional[CheckpointHandler]]:
-    logging.info(
-        '`_known_handlers` has been deprecated and should not be used.'
-    )
-    return dict(self._get_all_registered_and_unregistered_items_and_handlers())
+    self._metadata_store = checkpoint.metadata_store(enable_write=False)
 
   def _get_all_registered_and_unregistered_items_and_handlers(
       self,
@@ -629,7 +637,7 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
   ) -> atomicity_types.TemporaryPath:
     temporary_path_class = (
         self._temporary_path_class
-        or atomicity_defaults.get_default_temporary_path_class(directory)
+        or atomicity_defaults.get_item_default_temporary_path_class(directory)
     )
     tmp_item_dir = temporary_path_class.from_final(
         self._get_item_directory(directory, item_name),
@@ -669,10 +677,20 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
     self._current_temporary_paths = self._get_item_temporary_paths(
         directory, args
     )
-    await atomicity.create_all(
-        list(self._current_temporary_paths.values()),
-        multiprocessing_options=self._multiprocessing_options,
-    )
+    commit_futures = []
+    if self._async_options.create_directories_asynchronously:
+      commit_futures.append(
+          atomicity.create_all_async(
+              list(self._current_temporary_paths.values()),
+              completion_signals=_DIRECTORY_CREATION_SIGNALS,
+              multiprocessing_options=self._multiprocessing_options,
+          )
+      )
+    else:
+      await atomicity.create_all(
+          list(self._current_temporary_paths.values()),
+          multiprocessing_options=self._multiprocessing_options,
+      )
     save_ops = []
     for item_name, item_directory in self._current_temporary_paths.items():
       arg = args[item_name]
@@ -684,8 +702,10 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
         # Blocking save.
         handler.save(item_directory.get(), args=arg)
 
-    commit_futures = jax.tree.flatten(await asyncio.gather(*save_ops))[0]
-    return commit_futures or []
+    commit_futures.extend(
+        jax.tree.flatten(await asyncio.gather(*save_ops))[0] or []
+    )
+    return commit_futures
 
   def save(self, *args, **kwargs):
     """Saves synchronously."""
@@ -717,11 +737,7 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
     }
 
   def _existing_items(self, directory: epath.Path) -> List[str]:
-    return [
-        p.name
-        for p in directory.iterdir()
-        if p.is_dir()
-    ]
+    return [p.name for p in directory.iterdir() if p.is_dir()]
 
   def restore(
       self,
@@ -761,17 +777,29 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
 
     if args is None or not args.items():
       composite_args_items = {}
+      item_handlers = None
       for item_name in existing_items:
-        if item_name not in items_to_handlers:
-          raise KeyError(
-              f'Item "{item_name}" was found in the checkpoint, but could not'
-              ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
-              ' call `restore` with an appropriate `CheckpointArgs` subclass.'
-          )
-        handler = items_to_handlers[item_name]
         # Skip reserved items (they were not specifically requested).
         if item_name in RESERVED_ITEM_NAMES:
           continue
+
+        if item_name not in items_to_handlers:
+          item_handlers = (
+              item_handlers or self.metadata(directory).item_handlers
+          )
+          try:
+            # TODO(b/392622978): Support custom (local) handler registry.
+            handler = handler_type_registry.get_handler_type(
+                item_handlers[item_name]
+            )
+          except KeyError as e:
+            raise ValueError(
+                f'Item "{item_name}" was found in the checkpoint, but could not'
+                ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
+                ' call `restore` with an appropriate `CheckpointArgs` subclass.'
+            ) from e
+        else:
+          handler = items_to_handlers[item_name]
         if handler is None:
           raise ValueError(
               f'Item with name: "{item_name}" had an undetermined'
@@ -813,7 +841,124 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
       )
     return CompositeResults(**restored)
 
-  def metadata(self, directory: epath.Path) -> CompositeResults:
+  def _get_item_handlers(
+      self,
+      saved_metadata: checkpoint.StepMetadata,
+      item_names: list[str],
+      items_to_handlers: dict[str, CheckpointHandler | None],
+  ) -> dict[str, checkpoint.CheckpointHandlerTypeStr | None]:
+    if saved_metadata.item_handlers is not None:
+      assert isinstance(saved_metadata.item_handlers, dict)
+      # Keep relevant non-None handler typestrs from on-disk metadata.
+      item_handlers: dict[str, checkpoint.CheckpointHandlerTypeStr] = (
+          saved_metadata.item_handlers
+      )
+    else:
+      item_handlers: dict[str, checkpoint.CheckpointHandlerTypeStr] = {}
+    for item_name in item_names:
+      if items_to_handlers.get(item_name) is None:
+        logging.warning(
+            'Item "%s" was found in the checkpoint, but could not'
+            ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
+            ' call `restore` with an appropriate `CheckpointArgs` subclass.',
+            item_name,
+        )
+        # Don't overwrite if it was already set on disk.
+        if item_name not in item_handlers:
+          item_handlers[item_name] = None
+        continue
+
+      handler = items_to_handlers[item_name]
+      assert handler is not None
+      # If already set on disk (to non-None), don't overwrite.
+      if item_handlers.get(item_name) is None:
+        item_handlers[item_name] = handler.typestr()
+
+    return item_handlers
+
+  def _get_item_metadata(
+      self,
+      directory: epath.Path,
+      item_names: list[str],
+      items_to_handlers: dict[str, CheckpointHandler | None],
+  ) -> dict[str, Any]:
+    # item_metadata is not saved in StepMetadata, so we don't need to worry
+    # about reading it from disk.
+    item_metadata = {}
+    for item_name in item_names:
+      if items_to_handlers.get(item_name) is None:
+        logging.warning(
+            'Item "%s" was found in the checkpoint, but could not'
+            ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
+            ' call `restore` with an appropriate `CheckpointArgs` subclass.',
+            item_name,
+        )
+        item_metadata[item_name] = None
+        continue
+
+      handler = items_to_handlers[item_name]
+      assert handler is not None
+      item_metadata[item_name] = handler.metadata(
+          self._get_item_directory(directory, item_name)
+      )
+    return item_metadata
+
+  def _get_metadata_base(
+      self,
+      directory: epath.Path,
+      item_names: list[str],
+      get_item_metadata: bool = True,
+  ) -> StepMetadata:
+    """Base implementation for metadata handling.
+
+    Args:
+      directory: Path to the checkpoint.
+      item_names: List of item names to process.
+      get_item_metadata: whether to get item metadata,
+
+    Returns:
+        A tuple of item handlers and item metadata.
+    """
+    items_to_handlers = dict(
+        self._get_all_registered_and_unregistered_items_and_handlers()
+    )
+
+    serialized_metadata = self._metadata_store.read(
+        checkpoint.step_metadata_file_path(directory)
+    )
+    saved_metadata = step_metadata_serialization.deserialize(
+        serialized_metadata or {}
+    )
+
+    if get_item_metadata:
+      item_metadata = CompositeItemMetadata(
+          **self._get_item_metadata(directory, item_names, items_to_handlers)
+      )
+    else:
+      item_metadata = None
+
+    return dataclasses.replace(
+        saved_metadata,
+        item_handlers=self._get_item_handlers(
+            saved_metadata, item_names, items_to_handlers
+        ),
+        item_metadata=item_metadata,
+    )
+
+  def metadata_from_temporary_paths(
+      self, directory: epath.Path
+  ) -> StepMetadata:
+    """Metadata for each item in the temporary checkpoint."""
+    if not directory.exists():
+      raise FileNotFoundError(f'Directory does not exist: {directory}')
+
+    return self._get_metadata_base(
+        directory=directory,
+        item_names=list(self._current_temporary_paths.keys()),
+        get_item_metadata=False,
+    )
+
+  def metadata(self, directory: epath.Path) -> StepMetadata:
     """Metadata for each item in the checkpoint.
 
     This has much the same logic as `restore`, in the sense that it tries to
@@ -825,33 +970,36 @@ class CompositeCheckpointHandler(AsyncCheckpointHandler):
       directory: Path to the checkpoint.
 
     Returns:
-      CompositeResults
-    """
-    items_to_handlers = dict(
-        self._get_all_registered_and_unregistered_items_and_handlers()
-    )
-    existing_items = self._existing_items(directory)
+      StepMetadata
 
-    metadata = {}
-    for item_name in existing_items:
-      if (
-          item_name not in items_to_handlers
-          or items_to_handlers[item_name] is None
-      ):
-        logging.warning(
-            'Item "%s" was found in the checkpoint, but could not'
-            ' be restored. Please provide a `CheckpointHandlerRegistry`, or'
-            ' call `restore` with an appropriate `CheckpointArgs` subclass.',
-            item_name,
-        )
-        metadata[item_name] = None
-        continue
-      handler = items_to_handlers[item_name]
-      assert handler is not None
-      metadata[item_name] = handler.metadata(
-          self._get_item_directory(directory, item_name)
+    Raises:
+      FileNotFoundError: If the directory does not exist.
+    """
+    if not directory.exists():
+      raise FileNotFoundError(f'Directory does not exist: {directory}')
+
+    try:
+      existing_items = self._existing_items(directory)
+    except OSError:
+      existing_items = []
+      logging.warning(
+          'Failed to get existing items from directory %s. Will use items '
+          'provided during initialization.',
+          directory,
       )
-    return CompositeResults(**metadata)
+
+    for item_tmp_dir in self._current_temporary_paths.values():
+      tmp_dir_name = item_tmp_dir.get().name
+      if tmp_dir_name in existing_items:
+        raise ValueError(
+            f'Item "{tmp_dir_name}" was found in the checkpoint, but it is a '
+            'temporary item. Please call finalize() before metadata().'
+        )
+
+    return self._get_metadata_base(
+        directory=directory,
+        item_names=existing_items,
+    )
 
   def finalize(self, directory: epath.Path):
     if not self._current_temporary_paths:
