@@ -34,18 +34,20 @@ import jax
 import numpy as np
 from orbax.checkpoint import aggregate_handlers
 from orbax.checkpoint import checkpoint_args
-from orbax.checkpoint import future
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import transform_utils
 from orbax.checkpoint import utils
 from orbax.checkpoint._src import asyncio_utils
+from orbax.checkpoint._src.futures import future
 from orbax.checkpoint._src.handlers import async_checkpoint_handler
 from orbax.checkpoint._src.handlers import base_pytree_checkpoint_handler
+from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
 from orbax.checkpoint._src.metadata import empty_values
 from orbax.checkpoint._src.metadata import tree as tree_metadata
 from orbax.checkpoint._src.serialization import serialization
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
 from orbax.checkpoint._src.serialization import type_handlers
+from orbax.checkpoint._src.tree import types as tree_types
 from orbax.checkpoint._src.tree import utils as tree_utils
 import tensorstore as ts
 
@@ -268,6 +270,9 @@ def _get_restore_parameters(
         byte_limiter=byte_limiter,
         use_zarr3=use_zarr3,
         ts_context=ts_context,
+        # Skip raising array data missing error on this code path, since it
+        # almost exclusively handles legacy use cases.
+        raise_array_data_missing_error=False,
     )
 
   if transforms is None:
@@ -428,6 +433,7 @@ def _get_impl_save_args(
       save_args=args.save_args,
       ocdbt_target_data_file_size=args.ocdbt_target_data_file_size,
       enable_pinned_host_transfer=args.enable_pinned_host_transfer,
+      custom_metadata=args.custom_metadata,
   )
 
 
@@ -472,6 +478,9 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
       pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
           tree_metadata.PYTREE_METADATA_OPTIONS
       ),
+      array_metadata_validator: array_metadata_store_lib.Validator = (
+          array_metadata_store_lib.Validator()
+      ),
   ):
     """Creates PyTreeCheckpointHandler.
 
@@ -493,6 +502,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         specified, the global type handler registry will be used.
       handler_impl: Allows overriding the internal implementation.
       pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
+      array_metadata_validator: Validator for ArrayMetadata.
     """
     self._aggregate_handler = MsgpackHandler(
         primary_host=multiprocessing_options.primary_host,
@@ -515,6 +525,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         multiprocessing_options=multiprocessing_options,
         type_handler_registry=type_handler_registry,
         pytree_metadata_options=pytree_metadata_options,
+        array_metadata_validator=array_metadata_validator,
     )
     self._pytree_metadata_options = pytree_metadata_options
 
@@ -793,6 +804,11 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
     # `checkpoint_restore_args` has a structure relative to the checkpoint,
     # while `restore_args` remains structured relative to the output.
 
+    use_zarr3 = (
+        use_zarr3_metadata
+        if use_zarr3_metadata is not None
+        else self._use_zarr3
+    )
     param_infos, checkpoint_restore_args = _get_restore_parameters(
         directory,
         item,
@@ -802,9 +818,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         restore_args,
         self._pytree_metadata_options,
         transforms_default_to_original=transforms_default_to_original,
-        use_zarr3=use_zarr3_metadata
-        if use_zarr3_metadata is not None
-        else self._use_zarr3,
+        use_zarr3=use_zarr3,
     )
 
     if legacy_transform_fn is not None and transforms is not None:
@@ -982,7 +996,7 @@ class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
         use_zarr3,
     )
 
-  def metadata(self, directory: epath.Path) -> Optional[PyTree]:
+  def metadata(self, directory: epath.Path) -> tree_metadata.TreeMetadata:
     """Returns tree metadata.
 
     The result will be a PyTree matching the structure of the saved checkpoint.
@@ -1043,15 +1057,23 @@ class PyTreeSaveArgs(CheckpointArgs):
       indicates no maximum file size limit.  For best results, ensure
       chunk_byte_size is smaller than this value.  For more details, refer to
       https://google.github.io/tensorstore/kvstore/ocdbt/index.html#json-kvstore/ocdbt.target_data_file_size
-    enable_pinned_host_transfer: True by default. If False, disables transfer to
-      pinned host when copying from device to host, regardless of the presence
-      of pinned host memory.
+    enable_pinned_host_transfer: If False, disables transfer to pinned host when
+      copying from device to host, regardless of the presence of pinned host
+      memory.
+    custom_metadata: User-provided custom metadata. An arbitrary
+      JSON-serializable dictionary the user can use to store additional
+      information. The field is treated as opaque by Orbax.
   """
 
   item: PyTree
   save_args: Optional[PyTree] = None
   ocdbt_target_data_file_size: Optional[int] = None
-  enable_pinned_host_transfer: bool = True
+  enable_pinned_host_transfer: bool = False
+  custom_metadata: tree_types.JsonType | None = None
+
+  def __post_init__(self):
+    if isinstance(self.item, tree_metadata.TreeMetadata):
+      raise ValueError('Cannot save TreeMetadata.')
 
 
 @register_with_handler(PyTreeCheckpointHandler, for_restore=True)
@@ -1064,6 +1086,8 @@ class PyTreeRestoreArgs(CheckpointArgs):
       will infer the structure from the saved checkpoint. Transformations will
       not be run in this case. Necessary particularly in the case where the
       caller needs to restore the tree as a custom object.
+      `TreeMetadata` is also allowed as the tree used to
+      define the restored structure.
     restore_args: optional object containing additional arguments for
       restoration. It should be a PyTree matching the structure of `item`, or
       if `item` is not provided, then it should match the structure of the
@@ -1072,11 +1096,13 @@ class PyTreeRestoreArgs(CheckpointArgs):
       leaf as a certain type, a specific subclass of `RestoreArgs` may be
       required. `RestoreArgs` also provides the option to customize the
       restore type of an individual leaf.
+      `TreeMetadata` is also allowed as the `restore_args` tree.
     transforms: a PyTree of transformations that should be applied to the
       saved tree in order to obtain a final structure. The `transforms` tree
       structure should conceptually match that of `item`, but the use of
       regexes and implicit keys means that it does not need to match
       completely. See `transform_utils` for further information.
+      `TreeMetadata` is also allowed as the `transforms` tree.
     transforms_default_to_original: See transform_utils.apply_transformations.
     legacy_transform_fn: WARNING: NOT GENERALLY SUPPORTED. A function which
       accepts the `item` argument, a PyTree checkpoint structure and a PyTree
@@ -1090,3 +1116,11 @@ class PyTreeRestoreArgs(CheckpointArgs):
   transforms: Optional[PyTree] = None
   transforms_default_to_original: bool = True
   legacy_transform_fn: Optional[LegacyTransformFn] = None
+
+  def __post_init__(self):
+    if isinstance(self.item, tree_metadata.TreeMetadata):
+      self.item = self.item.tree
+    if isinstance(self.restore_args, tree_metadata.TreeMetadata):
+      self.restore_args = self.restore_args.tree
+    if isinstance(self.transforms, tree_metadata.TreeMetadata):
+      self.transforms = self.transforms.tree
