@@ -19,7 +19,7 @@ import concurrent.futures
 import dataclasses
 import json
 import threading
-from typing import Any, Protocol, TypeVar
+from typing import Any, Mapping, Protocol, TypeAlias, TypeVar
 
 from absl import logging
 from etils import epath
@@ -30,7 +30,10 @@ _STEP_METADATA_FILENAME = '_CHECKPOINT_METADATA'
 _ROOT_METADATA_FILENAME = '_ROOT_METADATA'
 _LEGACY_ROOT_METADATA_FILENAME = 'metadata'
 
-ItemMetadata = composite.Composite
+CompositeCheckpointHandlerTypeStrs: TypeAlias = Mapping
+CheckpointHandlerTypeStr = str
+CompositeItemMetadata = composite.Composite
+SingleItemMetadata = Any
 StepStatistics = step_statistics.SaveStepStatistics
 SerializedMetadata = TypeVar('SerializedMetadata', bound=dict[str, Any])
 
@@ -65,29 +68,32 @@ class StepMetadata:
   """Metadata of a checkpoint at step level (not per item).
 
   Attributes:
-    format: The checkpoint file format. Users should specify the format
-      explicitly when using something non-standard.
-    item_handlers: Map of item name to its checkpoint handler.
-    item_metadata: Map of item name to its metadata.
+    item_handlers: Map of item name to its checkpoint handler. Or a single
+      checkpoint handler for non composite checkpoints.
+    item_metadata: Map of item name to its metadata. Or a single metadata for
+      non composite checkpoints.
     metrics: User-provided metrics (accuracy, loss, etc.)
     performance_metrics: Performance metrics (time, memory, etc.)
     init_timestamp_nsecs: timestamp when uncommitted checkpoint was initialized.
       Specified as nano seconds since epoch. default=None.
     commit_timestamp_nsecs: commit timestamp of a checkpoint, specified as nano
       seconds since epoch. default=None.
-    custom: User-provided custom metadata.
+    custom_metadata: User-provided custom metadata. An arbitrary
+      JSON-serializable dictionary the user can use to store additional
+      information. The field is treated as opaque by Orbax.
   """
 
-  format: str | None = None
-  item_handlers: dict[str, str] = dataclasses.field(default_factory=dict)
-  item_metadata: ItemMetadata | None = None
+  item_handlers: (
+      dict[str, CheckpointHandlerTypeStr] | CheckpointHandlerTypeStr | None
+  ) = None
+  item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None
   metrics: dict[str, Any] = dataclasses.field(default_factory=dict)
   performance_metrics: StepStatistics = dataclasses.field(
       default_factory=StepStatistics
   )
   init_timestamp_nsecs: int | None = None
   commit_timestamp_nsecs: int | None = None
-  custom: dict[str, Any] = dataclasses.field(default_factory=dict)
+  custom_metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
 
 
 @dataclasses.dataclass
@@ -95,13 +101,14 @@ class RootMetadata:
   """Metadata of a checkpoint at root level (contains all steps).
 
   Attributes:
-    format: The checkpoint file format. Users should specify the format
-      explicitly when using something non-standard.
-    custom: User-provided custom metadata.
+    custom_metadata: User-provided custom metadata. An arbitrary
+      JSON-serializable dictionary the user can use to store additional
+      information. The field is treated as opaque by Orbax.
   """
 
-  format: str | None = None
-  custom: dict[str, Any] | None = dataclasses.field(default_factory=dict)
+  custom_metadata: dict[str, Any] | None = dataclasses.field(
+      default_factory=dict
+  )
 
 
 class MetadataStore(Protocol):
