@@ -29,8 +29,8 @@ StepMetadata = checkpoint.StepMetadata
 RootMetadata = checkpoint.RootMetadata
 MetadataStore = checkpoint.MetadataStore
 StepStatistics = step_statistics.SaveStepStatistics
-
-_SAMPLE_FORMAT = 'sample_format'
+CompositeItemMetadata = checkpoint.CompositeItemMetadata
+SingleItemMetadata = checkpoint.SingleItemMetadata
 
 
 class CheckpointMetadataTest(parameterized.TestCase):
@@ -69,7 +69,7 @@ class CheckpointMetadataTest(parameterized.TestCase):
       self,
       metadata_type: type[StepMetadata] | type[RootMetadata],
       metadata: checkpoint.SerializedMetadata,
-      item_metadata: checkpoint.ItemMetadata | None = None,
+      item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
       metrics: dict[str, Any] | None = None,
   ) -> StepMetadata | RootMetadata:
     if metadata_type == StepMetadata:
@@ -93,16 +93,15 @@ class CheckpointMetadataTest(parameterized.TestCase):
   def get_metadata(
       self,
       metadata_type: type[StepMetadata] | type[RootMetadata],
-      custom: dict[str, Any] = None,
+      custom_metadata: dict[str, Any] = None,
   ) -> StepMetadata | RootMetadata:
     """Returns a sample metadata object of `metadata_class`."""
-    if custom is None:
-      custom = {'a': 1}
+    if custom_metadata is None:
+      custom_metadata = {'a': 1}
     if metadata_type == StepMetadata:
       return StepMetadata(
-          format=_SAMPLE_FORMAT,
           item_handlers={'a': 'b'},
-          item_metadata={'a': None},
+          item_metadata=None,
           metrics={'a': 1},
           performance_metrics=StepStatistics(
               step=None,
@@ -112,12 +111,11 @@ class CheckpointMetadataTest(parameterized.TestCase):
           ),
           init_timestamp_nsecs=1,
           commit_timestamp_nsecs=1,
-          custom=custom,
+          custom_metadata=custom_metadata,
       )
     elif metadata_type == RootMetadata:
       return RootMetadata(
-          format=_SAMPLE_FORMAT,
-          custom=custom,
+          custom_metadata=custom_metadata,
       )
 
   def get_metadata_filename(
@@ -138,6 +136,20 @@ class CheckpointMetadataTest(parameterized.TestCase):
     elif metadata_type == RootMetadata:
       return checkpoint.root_metadata_file_path(path or self.directory)
 
+  def assertMetadataEqual(
+      self, a: StepMetadata | RootMetadata, b: StepMetadata | RootMetadata,
+  ):
+    if isinstance(a, StepMetadata):
+      self.assertEqual(a.item_handlers, b.item_handlers)
+      # ignore item_metadata
+      self.assertEqual(a.metrics, b.metrics)
+      self.assertEqual(a.performance_metrics, b.performance_metrics)
+      self.assertEqual(a.init_timestamp_nsecs, b.init_timestamp_nsecs)
+      self.assertEqual(a.commit_timestamp_nsecs, b.commit_timestamp_nsecs)
+      self.assertEqual(a.custom_metadata, b.custom_metadata)
+    elif isinstance(a, RootMetadata):
+      self.assertEqual(a.custom_metadata, b.custom_metadata)
+
   @parameterized.parameters(True, False)
   def test_read_unknown_path(self, blocking_write: bool):
     self.assertIsNone(
@@ -219,7 +231,7 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.write_metadata_store(blocking_write).read(
         file_path=self.get_metadata_file_path(metadata_class),
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
         metadata,
     )
@@ -247,7 +259,7 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.write_metadata_store(blocking_write).read(
         file_path=self.get_metadata_file_path(metadata_class),
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
         metadata,
     )
@@ -285,7 +297,6 @@ class CheckpointMetadataTest(parameterized.TestCase):
   ):
     self.write_metadata_store(blocking_write).update(
         file_path=self.get_metadata_file_path(metadata_class),
-        format=_SAMPLE_FORMAT,
         custom={'a': 1},
     )
 
@@ -294,46 +305,43 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.write_metadata_store(blocking_write).read(
         file_path=self.get_metadata_file_path(metadata_class),
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
         metadata_class(
-            format=_SAMPLE_FORMAT,
-            custom={'a': 1},
+            custom_metadata={'a': 1},
         ),
     )
 
   @parameterized.parameters(
       itertools.product(
           [True, False],
-          [StepMetadata, RootMetadata],
       )
   )
   def test_update_with_prior_data(
       self,
       blocking_write: bool,
-      metadata_class: type[StepMetadata] | type[RootMetadata],
   ):
-    metadata = metadata_class(format=_SAMPLE_FORMAT)
+    metadata = StepMetadata(init_timestamp_nsecs=1)
     self.write_metadata_store(blocking_write).write(
-        file_path=self.get_metadata_file_path(metadata_class),
+        file_path=self.get_metadata_file_path(StepMetadata),
         metadata=self.serialize_metadata(metadata),
     )
 
     self.write_metadata_store(blocking_write).update(
-        file_path=self.get_metadata_file_path(metadata_class),
+        file_path=self.get_metadata_file_path(StepMetadata),
         custom={'a': 1},
     )
 
     self.write_metadata_store(blocking_write).wait_until_finished()
 
     serialized_metadata = self.write_metadata_store(blocking_write).read(
-        file_path=self.get_metadata_file_path(metadata_class)
+        file_path=self.get_metadata_file_path(StepMetadata)
     )
-    self.assertEqual(
-        self.deserialize_metadata(metadata_class, serialized_metadata),
-        metadata_class(
-            format=_SAMPLE_FORMAT,
-            custom={'a': 1},
+    self.assertMetadataEqual(
+        self.deserialize_metadata(StepMetadata, serialized_metadata),
+        StepMetadata(
+            init_timestamp_nsecs=1,
+            custom_metadata={'a': 1},
         ),
     )
 
@@ -355,7 +363,6 @@ class CheckpointMetadataTest(parameterized.TestCase):
     )
     self.write_metadata_store(blocking_write).update(
         file_path=self.get_metadata_file_path(metadata_class),
-        format=_SAMPLE_FORMAT,
         blah=2,
     )
 
@@ -364,11 +371,10 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.write_metadata_store(blocking_write).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
         metadata_class(
-            format=_SAMPLE_FORMAT,
-            custom={'blah': 2},
+            custom_metadata={'blah': 2},
         ),
     )
 
@@ -415,14 +421,14 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.read_metadata_store(blocking_write=True).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
         self.get_metadata(metadata_class),
     )
 
     # write validations
     serialized_metadata = self.serialize_metadata(
-        self.get_metadata(metadata_class, custom={'a': 2})
+        self.get_metadata(metadata_class, custom_metadata={'a': 2})
     )
     self.write_metadata_store(blocking_write=False).write(
         file_path=self.get_metadata_file_path(metadata_class),
@@ -434,16 +440,16 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.read_metadata_store(blocking_write=False).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
-        self.get_metadata(metadata_class, custom={'a': 2}),
+        self.get_metadata(metadata_class, custom_metadata={'a': 2}),
     )
     serialized_metadata = self.write_metadata_store(blocking_write=False).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
-        self.get_metadata(metadata_class, custom={'a': 2}),
+        self.get_metadata(metadata_class, custom_metadata={'a': 2}),
     )
 
     # update validations
@@ -455,16 +461,16 @@ class CheckpointMetadataTest(parameterized.TestCase):
     serialized_metadata = self.read_metadata_store(blocking_write=False).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
-        self.get_metadata(metadata_class, custom={'a': 3}),
+        self.get_metadata(metadata_class, custom_metadata={'a': 3}),
     )
     serialized_metadata = self.write_metadata_store(blocking_write=False).read(
         file_path=self.get_metadata_file_path(metadata_class)
     )
-    self.assertEqual(
+    self.assertMetadataEqual(
         self.deserialize_metadata(metadata_class, serialized_metadata),
-        self.get_metadata(metadata_class, custom={'a': 3}),
+        self.get_metadata(metadata_class, custom_metadata={'a': 3}),
     )
 
   @parameterized.parameters(StepMetadata, RootMetadata)
@@ -494,7 +500,6 @@ class CheckpointMetadataTest(parameterized.TestCase):
     _ = pickle.dumps(self.read_metadata_store(blocking_write))
 
   @parameterized.parameters(
-      ({'format': int()},),
       ({'custom': list()},),
       ({'custom': {int(): None}},),
   )
@@ -505,11 +510,8 @@ class CheckpointMetadataTest(parameterized.TestCase):
       self.deserialize_metadata(RootMetadata, wrong_metadata)
 
   @parameterized.parameters(
-      ({'format': int()},),
       ({'item_handlers': list()},),
       ({'item_handlers': {int(): None}},),
-      ({'item_metadata': list()},),
-      ({'item_metadata': {int(): None}},),
       ({'metrics': list()},),
       ({'metrics': {int(): None}},),
       ({'performance_metrics': list()},),
@@ -526,44 +528,12 @@ class CheckpointMetadataTest(parameterized.TestCase):
     with self.assertRaises(ValueError):
       self.deserialize_metadata(StepMetadata, wrong_metadata)
 
-  @parameterized.parameters(
-      (
-          RootMetadata(custom={'a': None}),
-          {'custom': {'a': None}}
-      ),
-      (
-          RootMetadata(format=_SAMPLE_FORMAT),
-          {'format': _SAMPLE_FORMAT}
-      ),
-      (
-          StepMetadata(format=_SAMPLE_FORMAT),
-          {'format': _SAMPLE_FORMAT},
-      ),
-      (
-          StepMetadata(item_handlers={'a': 'a_handler'}),
-          {'item_handlers': {'a': 'a_handler'}},
-      ),
-      (
-          StepMetadata(custom={'blah': 123}),
-          {'custom': {'blah': 123}},
-      ),
-  )
-  def test_only_serialize_non_default_metadata_values(
-      self,
-      metadata: StepMetadata | RootMetadata,
-      expected_serialized_metadata: dict[str, Any],
-  ):
-    self.assertEqual(
-        self.serialize_metadata(metadata), expected_serialized_metadata
-    )
-
   @parameterized.parameters(StepMetadata, RootMetadata)
   def test_unknown_key_in_metadata(
       self, metadata_class: type[StepMetadata] | type[RootMetadata],
   ):
     metadata = metadata_class(
-        format=_SAMPLE_FORMAT,
-        custom={'a': 1},
+        custom_metadata={'a': 1},
     )
     serialized_metadata = self.serialize_metadata(metadata)
     serialized_metadata['blah'] = 2
@@ -572,6 +542,158 @@ class CheckpointMetadataTest(parameterized.TestCase):
     ):
       self.deserialize_metadata(metadata_class, serialized_metadata)
 
+  @parameterized.named_parameters(
+      ('single', 'a_handler'),
+      ('composite', {'a': 'a_handler'}),
+  )
+  def test_deserialize_item_handlers(self, item_handlers):
+    serialized_metadata = {'item_handlers': item_handlers}
+    metadata = self.deserialize_metadata(StepMetadata, serialized_metadata)
+    self.assertEqual(metadata.item_handlers, item_handlers)
+
+  @parameterized.named_parameters(
+      ('single', 'a'),
+      ('composite', {'A': 'a'}),
+  )
+  def test_deserialize_item_metadata_with_item_metadata_kwarg(
+      self, item_metadata
+  ):
+    metadata = self.deserialize_metadata(
+        StepMetadata, {}, item_metadata=item_metadata,
+    )
+    self.assertEqual(metadata.item_metadata, item_metadata)
+
+  @parameterized.parameters(
+      ('metrics', 1, dict, int),
+  )
+  def test_validate_type_wrong_type(
+      self, kwarg_name, kwarg_value, expected_type, wrong_type
+  ):
+    with self.assertRaisesRegex(
+        ValueError,
+        f'Object must be of type {expected_type}, got {wrong_type}.',
+    ):
+      self.deserialize_metadata(StepMetadata, {}, **{kwarg_name: kwarg_value})
+
+  @parameterized.parameters(
+      ({'metrics': 1}, dict, int),
+      ({'performance_metrics': 1}, dict, int),
+      ({'init_timestamp_nsecs': 'a'}, int, str),
+      ({'commit_timestamp_nsecs': 'a'}, int, str),
+      ({'custom': 1}, dict, int),
+  )
+  def test_validate_field_wrong_type(
+      self, step_metadata, expected_field_type, wrong_field_type
+  ):
+    with self.assertRaisesRegex(
+        ValueError,
+        f'Object must be of type {expected_field_type}, got {wrong_field_type}',
+    ):
+      self.deserialize_metadata(StepMetadata, step_metadata)
+
+  @parameterized.parameters(
+      ({'item_handlers': 1}, int),
+  )
+  def test_validate_field_sequence_wrong_type(
+      self, step_metadata, wrong_field_type
+  ):
+    with self.assertRaisesRegex(
+        ValueError,
+        f'Object must be any one of types '
+        r'\[.+]'
+        f', got {wrong_field_type}.',
+    ):
+      self.deserialize_metadata(StepMetadata, step_metadata)
+
+  @parameterized.parameters(
+      ({'item_handlers': {1: 'a'}}, str, int),
+      ({'metrics': {1: 'a'}}, str, int),
+      ({'performance_metrics': {1: 1.0}}, str, int),
+      ({'custom': {1: 'a'}}, str, int),
+  )
+  def test_validate_dict_entry_wrong_key_type(
+      self, step_metadata, expected_key_type, wrong_key_type
+  ):
+    with self.assertRaisesRegex(
+        ValueError,
+        f'Object must be of type {expected_key_type}, got {wrong_key_type}.',
+    ):
+      self.deserialize_metadata(StepMetadata, step_metadata)
+
+  @parameterized.parameters(
+      ({'performance_metrics': {'a': 1}}, float, int),
+  )
+  def test_validate_dict_entry_wrong_value_type(
+      self, step_metadata, expected_value_type, wrong_value_type
+  ):
+    with self.assertRaisesRegex(
+        ValueError,
+        f'Object must be of type {expected_value_type}, got {wrong_value_type}',
+    ):
+      self.deserialize_metadata(StepMetadata, step_metadata)
+
+  @parameterized.parameters(
+      ({'item_handlers': {'a': 'b'}},),
+      ({'performance_metrics': {'a': 1.0}},),
+      ({'custom_metadata': {'a': 1}, 'init_timestamp_nsecs': 1},),
+  )
+  def test_serialize_for_update_valid_kwargs(
+      self, kwargs: dict[str, Any]
+  ):
+    expected_kwargs = kwargs.copy()
+    if 'custom_metadata' in expected_kwargs:
+      expected_kwargs['custom'] = expected_kwargs.pop('custom_metadata')
+    self.assertEqual(
+        step_metadata_serialization.serialize_for_update(**kwargs),
+        expected_kwargs,
+    )
+
+  @parameterized.parameters(
+      ({'item_handlers': list()},),
+      ({'item_handlers': {int(): None}},),
+      ({'metrics': list()},),
+      ({'metrics': {int(): None}},),
+      ({'performance_metrics': list()},),
+      ({'init_timestamp_nsecs': float()},),
+      ({'commit_timestamp_nsecs': float()},),
+      ({'custom_metadata': list()},),
+      ({'custom_metadata': {int(): None}},),
+  )
+  def test_serialize_for_update_wrong_types(
+      self, kwargs: dict[str, Any]
+  ):
+    with self.assertRaises(ValueError):
+      step_metadata_serialization.serialize_for_update(**kwargs)
+
+  def test_serialize_for_update_with_unknown_kwargs(self):
+    with self.assertRaisesRegex(
+        ValueError, 'Provided metadata contains unknown key blah'
+    ):
+      step_metadata_serialization.serialize_for_update(
+          custom_metadata={'a': 1},
+          blah=123,
+      )
+
+    with self.assertRaisesRegex(
+        ValueError, 'Provided metadata contains unknown key custom'
+    ):
+      step_metadata_serialization.serialize_for_update(
+          custom={'a': 1},
+      )
+
+  def test_serialize_for_update_performance_metrics_only_float(self):
+    self.assertEqual(
+        step_metadata_serialization.serialize_for_update(
+            performance_metrics=StepStatistics(
+                step=1,
+                event_type='save',
+                reached_preemption=False,
+                preemption_received_at=1.0,
+            )
+        ),
+        {'performance_metrics': {'preemption_received_at': 1.0}},
+    )
+
 
 if __name__ == '__main__':
   absltest.main()
