@@ -91,7 +91,8 @@ def initialize_distributed_to_device_ids():
     ]
   assert None not in results
   _DISTRIBUTED_TO_DEVICE_IDS = results
-  logging.info(
+  logging.vlog(
+      1,
       '[process=%s][thread=%s] distributed_to_device_ids: %s',
       own_distributed_id,
       threading.current_thread().name,
@@ -124,7 +125,8 @@ def initialize_runtime_to_distributed_ids():
   for key, distributed_id in ids:
     runtime_id = int(key.split('/')[-1])
     _RUNTIME_TO_DISTRIBUTED_ID[runtime_id] = int(distributed_id)
-  logging.info(
+  logging.vlog(
+      1,
       '[process=%s][thread=%s] runtime_to_distributed_id: %s',
       process_index(),
       threading.current_thread().name,
@@ -160,6 +162,11 @@ def should_skip_process_sync(processes: Optional[Set[int]] = None) -> bool:
   return False
 
 
+def is_jax_distributed_client_initialized() -> bool:
+  """Returns True if the JAX distributed client is initialized."""
+  return jax._src.distributed.global_state.client is not None  # pylint: disable=protected-access
+
+
 def get_jax_distributed_client():
   client = jax._src.distributed.global_state.client  # pylint: disable=protected-access
   if client is None:
@@ -222,7 +229,8 @@ def get_barrier_sync_fn(
 
   def _fn(*, key: str, timeout_ms: int) -> None:
     key = _unique_barrier_key(key)
-    logging.info(
+    logging.vlog(
+        1,
         '[process=%s][thread=%s] Waiting at barrier: %s',
         process_index(),
         threading.current_thread().name,
@@ -239,7 +247,8 @@ def get_barrier_sync_fn(
           barrier_processes,
       )
       client.wait_at_barrier(key, timeout_ms, process_ids=barrier_processes)
-    logging.info(
+    logging.vlog(
+        1,
         '[process=%s][thread=%s] Done waiting at barrier: %s',
         process_index(),
         threading.current_thread().name,
@@ -290,7 +299,8 @@ def sync_global_processes(
       provided, a default implementation is used.
   """
   if should_skip_process_sync(processes):
-    logging.info(
+    logging.vlog(
+        1,
         '[process=%s][thread=%s] Skipping global process sync, barrier'
         ' name: %s',
         process_index(),
@@ -303,19 +313,21 @@ def sync_global_processes(
   # Temporarily default to existing behavior to minimize risk of breakage.
   if processes is None:
     key = _unique_barrier_key(name)
-    logging.info(
-        '[process=%s][thread=%s] Waiting with jax/sync_global_devices("%s")',
+    logging.vlog(
+        1,
+        '[process=%s][thread=%s] Begin jax/sync_global_devices("%s")',
         process_index(),
         threading.current_thread().name,
         key,
     )
     multihost_utils.sync_global_devices(key)
-    logging.info(
-        '[process=%s][thread=%s] Done waiting with'
-        ' jax/sync_global_devices("%s")',
+    logging.vlog(
+        1,
+        '[process=%s][thread=%s] Done jax/sync_global_devices("%s"): %s secs',
         process_index(),
         threading.current_thread().name,
         key,
+        (time.time() - sync_start_time),
     )
   else:
     barrier_sync_fn = barrier_sync_fn or get_barrier_sync_fn(
