@@ -27,9 +27,9 @@ from jax import dtypes as _dtypes
 from jax.experimental import layout
 import jax.numpy as jnp
 import numpy as np
-from orbax.checkpoint import future
 from orbax.checkpoint import test_utils
 from orbax.checkpoint._src import asyncio_utils
+from orbax.checkpoint._src.futures import future
 from orbax.checkpoint._src.serialization import serialization
 from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
 import tensorstore as ts
@@ -605,15 +605,12 @@ class CheckpointTest(parameterized.TestCase):
     s = NamedSharding(mesh, P('x', 'y'))
     arr = jax.device_put(np_inp, s)
 
-    out_layout = (
-        jax.jit(lambda x: x.T, out_shardings=Layout(DLL.AUTO))
-        .lower(arr)
-        .compile()
-        .output_layouts
-    )
-    self.assertEqual(
-        arr.layout.device_local_layout.major_to_minor,
-        out_layout.device_local_layout.major_to_minor[::-1],
+    out_layout = Layout(
+        device_local_layout=DLL(
+            arr.layout.device_local_layout.major_to_minor[::-1],
+            arr.layout.device_local_layout._tiling,
+        ),
+        sharding=arr.sharding,
     )
 
     ckpt_dir = pathlib.Path(self.create_tempdir('ckpt').full_path)
