@@ -35,10 +35,10 @@ import jax
 from jax.experimental import layout
 import jax.numpy as jnp
 import numpy as np
-from orbax.checkpoint import future
-from orbax.checkpoint._src import asyncio_utils
 from orbax.checkpoint._src.arrays import subchunking
 from orbax.checkpoint._src.arrays import types as arrays_types
+from orbax.checkpoint._src.futures import future
+from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
 from orbax.checkpoint._src.metadata import empty_values
 from orbax.checkpoint._src.metadata import sharding as sharding_metadata
 from orbax.checkpoint._src.metadata import value as value_metadata
@@ -57,6 +57,7 @@ SaveArgs: TypeAlias = types.SaveArgs
 RestoreArgs: TypeAlias = types.RestoreArgs
 TypeHandler: TypeAlias = types.TypeHandler
 TypeHandlerRegistry: TypeAlias = types.TypeHandlerRegistry
+Pytree = Any
 
 Layout = layout.Layout
 Shape = arrays_types.Shape
@@ -202,7 +203,7 @@ def get_json_tspec_write(
   return tspec
 
 
-def _build_array_tspec_write(
+def _build_array_write_spec(
     info: types.ParamInfo,
     arg: Optional[types.SaveArgs] = None,
     *,
@@ -212,8 +213,8 @@ def _build_array_tspec_write(
     use_ocdbt: bool,
     process_index: Optional[Union[int, str]] = None,
     metadata_key: Optional[str] = None,
-) -> ts_utils.JsonSpec:
-  """Gets Tensorstore spec for writing."""
+) -> ts_utils.ArrayWriteSpec:
+  """Gets ArrayWriteSpec for writing."""
   if info.path is None:
     raise ValueError('Must construct serialization path.')
   parent_dir = info.parent_dir
@@ -228,26 +229,13 @@ def _build_array_tspec_write(
       dtype=dtype,
       target_dtype=(arg.dtype if arg is not None else None),
       chunk_byte_size=(arg.chunk_byte_size if arg is not None else None),
+      shard_axes=(arg.shard_axes if arg is not None else tuple()),
       use_zarr3=info.use_zarr3,
       use_ocdbt=use_ocdbt,
       process_id=process_index,
       ocdbt_target_data_file_size=info.ocdbt_target_data_file_size,
       metadata_key=metadata_key,
-  ).json
-
-
-class _CommitFuture(future.Future):
-  """Represents the result of a background commit."""
-
-  def __init__(self, coro, name: Optional[str] = None):
-    self._t = future.ThreadRaisingException(
-        name=name,
-        target=lambda: asyncio_utils.run_sync(coro),
-    )
-    self._t.start()
-
-  def result(self, timeout: Optional[int] = None) -> Any:
-    return self._t.join(timeout=timeout)
+  )
 
 
 def check_input_arguments(*args):
@@ -261,6 +249,12 @@ def check_input_arguments(*args):
       raise ValueError('Found input args with mismatched lengths.')
 
 
+def check_array_values(values: Sequence[Union[jax.Array, np.ndarray]]):
+  for v in values:
+    if v.size == 0:
+      raise ValueError('Cannot save arrays with zero size.')
+
+
 async def _validate_params(
     directory: epath.Path,
     ts_context: ts.Context,
@@ -344,7 +338,8 @@ async def _validate_params(
         )
 
   unique = with_zarray | without_zarray
-  logging.info(
+  logging.vlog(
+      1,
       '[process=%s][thread=%s] Validating params in TensorStore KvStore.',
       process_index,
       current_thread_name,
@@ -503,7 +498,8 @@ def _array_metadata_from_tensorstore(
       dtype=jnp.dtype(t.dtype.name),
       sharding=sharding,
       storage=value_metadata.StorageMetadata(
-          chunk_shape=t.chunk_layout.read_chunk_template.shape
+          chunk_shape=t.chunk_layout.read_chunk_template.shape,
+          write_shape=info.write_shape,
       ),
   )
 
@@ -540,16 +536,16 @@ class NumpyHandler(types.TypeHandler):
     # managed by this controller.
     self._override_ocdbt_process_id: Optional[str] = None
 
-  def _get_json_tspec_write(
+  def _get_array_write_spec(
       self,
       info: types.ParamInfo,
       value: np.ndarray,
       use_ocdbt: bool,
       process_index: Optional[Union[int, str]] = None,
       arg: Optional[types.SaveArgs] = None,
-  ) -> Dict[str, Any]:
-    """Gets Tensorstore spec for writing."""
-    return _build_array_tspec_write(
+  ) -> ts_utils.ArrayWriteSpec:
+    """Gets ArrayWriteSpec for writing."""
+    return _build_array_write_spec(
         info=info,
         arg=arg,
         global_shape=value.shape,
@@ -612,7 +608,7 @@ class NumpyHandler(types.TypeHandler):
     """Serializes numpy arrays in a background thread."""
     write_coros = []
     for value, info, arg in zip(values, infos, args):
-      tspec = self._get_json_tspec_write(
+      array_write_spec = self._get_array_write_spec(
           info,
           value,
           use_ocdbt=info.is_ocdbt_checkpoint,
@@ -622,6 +618,7 @@ class NumpyHandler(types.TypeHandler):
           ),
           arg=arg,
       )
+      tspec = array_write_spec.json
       if logging.vlog_is_on(1):
         logging.vlog(1, 'tspec = %s', tspec)
         logging.vlog(1, 'infos = %s', info)
@@ -640,11 +637,12 @@ class NumpyHandler(types.TypeHandler):
     """Uses Tensorstore to serialize a numpy array."""
     args = args or [types.SaveArgs()] * len(values)
     check_input_arguments(values, infos, args)
+    check_array_values(values)
     if logging.vlog_is_on(1):
       _print_ts_debug_data(self._metadata_key, infos)
     copied_values = [copy.deepcopy(v) for v in values]
     return [
-        _CommitFuture(
+        future.CommitFutureAwaitingContractedSignals(
             self._background_serialize(copied_values, infos, args),
             name='np_type_handler',
         )
@@ -834,6 +832,29 @@ class SingleReplicaArrayRestoreArgs(ArrayRestoreArgs):
   single_replica_sharding: Optional[jax.sharding.NamedSharding] = None
 
 
+JAX_ARRAY_TYPE_STR = 'jax.Array'
+
+
+def represents_jax_array(param_info: types.ParamInfo) -> bool:
+  """Returns True if the param_info represents a jax.Array."""
+  assert (
+      param_info.value_typestr is not None
+  ), f'ParamInfo.value_typestr cannot be None: {param_info}'
+  return param_info.value_typestr == JAX_ARRAY_TYPE_STR
+
+
+def any_jax_array_param_info(param_infos: Pytree) -> types.ParamInfo | None:
+  """Returns any jax.Array param_info in the PyTree, or None."""
+  return jax.tree_util.tree_reduce(
+      lambda found_jax_array, param_info: (
+          found_jax_array
+          or (param_info if represents_jax_array(param_info) else None)
+      ),
+      tree=param_infos,
+      initializer=None,
+  )
+
+
 class ArrayHandler(types.TypeHandler):
   """An implementation of TypeHandler for jax.Array."""
 
@@ -842,8 +863,9 @@ class ArrayHandler(types.TypeHandler):
       metadata_key: Optional[str] = None,
       primary_host: Optional[int] = 0,
       replica_id: Optional[int] = 0,
-      use_replica_parallel: bool = True,
+      use_replica_parallel: bool = False,
       enable_write_sharding_file: bool = True,
+      array_metadata_store: array_metadata_store_lib.Store | None = None,
   ):
     """Constructor.
 
@@ -858,19 +880,25 @@ class ArrayHandler(types.TypeHandler):
       use_replica_parallel: Whether to parallelize saving across replicas.
       enable_write_sharding_file: whether to write sharding file, defaults to
         True.
+      array_metadata_store: Store to manage per host ArrayMetadata. To disable
+        ArrayMetadata persistence, set it to None.
     """
     self._metadata_key = metadata_key
     self._primary_host = primary_host
     self._replica_id = replica_id
     self._enable_write_sharding_file = enable_write_sharding_file
     self._use_replica_parallel = use_replica_parallel
+    self._array_metadata_store = array_metadata_store
 
-    logging.info(
-        'Created `ArrayHandler` with primary_host=%s, replica_id=%s,'
-        ' use_replica_parallel=%s',
+    logging.vlog(
+        1,
+        'Created `%s` with primary_host=%s, replica_id=%s,'
+        ' use_replica_parallel=%s, array_metadata_store=%s',
+        self.__class__.__qualname__,
         self._primary_host,
         self._replica_id,
         self._use_replica_parallel,
+        self._array_metadata_store,
     )
 
     if self._primary_host is None and jax.__version_info__ <= (0, 4, 25):  # pylint:disable=unreachable
@@ -878,7 +906,7 @@ class ArrayHandler(types.TypeHandler):
           'Setting `primary_host` to None requires JAX version > 0.4.25.'
       )
 
-  def _get_json_tspec_write(
+  def _get_array_write_spec(
       self,
       info: types.ParamInfo,
       value: replica_slices.ReplicaSlices,
@@ -886,9 +914,9 @@ class ArrayHandler(types.TypeHandler):
       use_ocdbt: bool,
       process_index: Optional[Union[int, str]] = None,
       arg: Optional[types.SaveArgs] = None,
-  ) -> Dict[str, Any]:
-    """Gets Tensorstore spec for writing."""
-    return _build_array_tspec_write(
+  ) -> ts_utils.ArrayWriteSpec:
+    """Gets ArrayWriteSpec for writing."""
+    return _build_array_write_spec(
         info=info,
         arg=arg,
         global_shape=value.global_shape,
@@ -913,7 +941,7 @@ class ArrayHandler(types.TypeHandler):
     )
 
   def typestr(self) -> str:
-    return 'jax.Array'
+    return JAX_ARRAY_TYPE_STR
 
   async def metadata(
       self, infos: Sequence[types.ParamInfo]
@@ -1011,6 +1039,7 @@ class ArrayHandler(types.TypeHandler):
     write_coros = []
     sharding_metadata_txn = ts.Transaction()
     ocdbt_transaction: Optional[ts.Transaction] = None
+    array_metadatas = []
     for value, info, arg in zip(values, infos, args):
       # The byte_limiter can't be used with a transaction, because awaiting the
       # `write` only waits until the in-memory transaction state reflects the
@@ -1019,13 +1048,14 @@ class ArrayHandler(types.TypeHandler):
       if info.is_ocdbt_checkpoint and info.byte_limiter is None:
         if ocdbt_transaction is None:
           ocdbt_transaction = ts.Transaction(atomic=True)
-      tspec = self._get_json_tspec_write(
+      array_write_spec = self._get_array_write_spec(
           info,
           value,
           use_ocdbt=info.is_ocdbt_checkpoint,
           process_index=get_process_index_for_subdir(info.is_ocdbt_checkpoint),
           arg=arg,
       )
+      tspec = array_write_spec.json
       ts_context = info.ts_context
       write_coros.append(
           serialization.async_serialize_from_host(
@@ -1043,7 +1073,15 @@ class ArrayHandler(types.TypeHandler):
                 value.sharding, info, sharding_metadata_txn
             )
         )
-
+      array_metadatas.append(array_write_spec.metadata)
+    if self._array_metadata_store is not None:
+      write_coros.append(
+          self._array_metadata_store.write(
+              checkpoint_dir=infos[0].parent_dir,
+              array_metadatas=array_metadatas,
+              process_index=multihost.process_index(),
+          )
+      )
     await asyncio.gather(*write_coros)
     await sharding_metadata_txn.commit_async()
     if ocdbt_transaction is not None:
@@ -1071,6 +1109,7 @@ class ArrayHandler(types.TypeHandler):
         )
     args = args or [types.SaveArgs()] * len(values)
     check_input_arguments(values, infos, args)
+    check_array_values(values)
 
     assert all([info.enable_pinned_host_transfer for info in infos]) or all(
         [not info.enable_pinned_host_transfer for info in infos]
@@ -1085,7 +1124,7 @@ class ArrayHandler(types.TypeHandler):
     )
 
     return [
-        _CommitFuture(
+        future.CommitFutureAwaitingContractedSignals(
             self._background_serialize(values_on_host, infos, args),
             name='array_type_handler',
         )
@@ -1175,8 +1214,9 @@ class ArrayHandler(types.TypeHandler):
       tspec = get_cast_tspec_deserialize(tspec, arg)
       if logging.vlog_is_on(1):
         logging.vlog(1, 'tspec = %s', tspec)
-        logging.vlog(1, 'infos = %s', infos)
-        logging.vlog(1, 'args = %s', args)
+        logging.vlog(1, 'info = %s', info)
+        logging.vlog(1, 'arg = %s', arg)
+        logging.vlog(1, 'sharding = %s', sharding)
       deserialize_ops += [
           serialization.async_deserialize(
               sharding,
@@ -1195,9 +1235,11 @@ class ArrayHandler(types.TypeHandler):
       for a in ret:
         logging.vlog(
             1,
-            'restored jax.Array.shape = %s, jax.array.dtype = %s',
+            'restored jax.Array.shape = %s, jax.array.dtype = %s,'
+            ' jax.array.layout + %s',
             a.shape,
             a.dtype,
+            a.layout,
         )
       _print_ts_debug_data(self._metadata_key, infos)
 
@@ -1527,7 +1569,7 @@ class StringHandler(types.TypeHandler):
     del args
     # Copy is not needed since strings are passed by value.
     return [
-        _CommitFuture(
+        future.CommitFutureAwaitingContractedSignals(
             self._background_serialize(values, infos),
             name='string_type_handler',
         )
@@ -1647,17 +1689,6 @@ class _TypeHandlerRegistryImpl(types.TypeHandlerRegistry):
       return False
 
 
-GLOBAL_TYPE_HANDLER_REGISTRY = _TypeHandlerRegistryImpl(
-    (int, ScalarHandler()),
-    (float, ScalarHandler()),
-    (bytes, ScalarHandler()),
-    (np.number, ScalarHandler()),
-    (np.ndarray, NumpyHandler()),
-    (jax.Array, ArrayHandler()),
-    (str, StringHandler()),
-)
-
-
 def create_type_handler_registry(
     *handlers: Tuple[Any, types.TypeHandler]
 ) -> types.TypeHandlerRegistry:
@@ -1673,6 +1704,22 @@ def create_type_handler_registry(
   return _TypeHandlerRegistryImpl(*handlers)
 
 
+_DEFAULT_TYPE_HANDLERS = tuple([
+    (int, ScalarHandler()),
+    (float, ScalarHandler()),
+    (bytes, ScalarHandler()),
+    (np.number, ScalarHandler()),
+    (np.ndarray, NumpyHandler()),
+    (jax.Array, ArrayHandler()),
+    (str, StringHandler()),
+])
+
+
+GLOBAL_TYPE_HANDLER_REGISTRY = create_type_handler_registry(
+    *_DEFAULT_TYPE_HANDLERS
+)
+
+
 def register_type_handler(
     ty: Any,
     handler: types.TypeHandler,
@@ -1710,6 +1757,11 @@ def has_type_handler(ty: Any) -> bool:
   return GLOBAL_TYPE_HANDLER_REGISTRY.has(ty)
 
 
+def supported_types() -> list[Any]:
+  """Returns the default list of supported types."""
+  return [ty for ty, _ in _DEFAULT_TYPE_HANDLERS]
+
+
 def register_standard_handlers_with_options(**kwargs):
   """Re-registers a select set of handlers with the given options.
 
