@@ -34,6 +34,7 @@ from orbax.checkpoint import args as args_lib
 from orbax.checkpoint import checkpoint_args
 from orbax.checkpoint import options as options_lib
 from orbax.checkpoint import utils
+from orbax.checkpoint._src.checkpoint_managers import save_decision_policy as save_decision_policy_lib
 from orbax.checkpoint._src.checkpointers import abstract_checkpointer
 from orbax.checkpoint._src.checkpointers import async_checkpointer
 from orbax.checkpoint._src.checkpointers import checkpointer as checkpointer_lib
@@ -43,6 +44,7 @@ from orbax.checkpoint._src.handlers import handler_registration
 from orbax.checkpoint._src.handlers import json_checkpoint_handler
 from orbax.checkpoint._src.handlers import proto_checkpoint_handler
 from orbax.checkpoint._src.metadata import checkpoint
+from orbax.checkpoint._src.metadata import checkpoint_info
 from orbax.checkpoint._src.metadata import root_metadata_serialization
 from orbax.checkpoint._src.multihost import multihost
 from orbax.checkpoint._src.path import atomicity_types
@@ -55,10 +57,12 @@ from orbax.checkpoint.logging import step_statistics
 from typing_extensions import Self  # for Python version < 3.11
 
 
+
 PyTree = Any
 CheckpointDirs = Tuple[str, str]
 SaveParams = Mapping[str, Any]
 RestoreParams = SaveParams
+CheckpointInfo = checkpoint_info.CheckpointInfo
 AbstractCheckpointer = abstract_checkpointer.AbstractCheckpointer
 CheckpointersDict = Mapping[str, AbstractCheckpointer]
 AbstractCheckpointManager = (
@@ -157,6 +161,52 @@ class _FinalizeThread(threading.Thread):
       raise self.exception
 
 
+@dataclasses.dataclass
+class _ShouldSaveFnPolicy(save_decision_policy_lib.SaveDecisionPolicy):
+  """A policy that uses a user-provided should_save_fn."""
+
+  should_save_fn: Callable[[int, Optional[int]], bool]
+
+  def should_save(
+      self,
+      step: save_decision_policy_lib.StepInfo,
+      previous_steps: Sequence[save_decision_policy_lib.StepInfo],
+      *,
+      context: save_decision_policy_lib.DecisionContext,
+  ) -> bool:
+    return self.should_save_fn(
+        step.step, previous_steps[-1].step if previous_steps else None
+    )
+
+
+def _get_default_save_decision_policy(
+    options: CheckpointManagerOptions,
+) -> save_decision_policy_lib.SaveDecisionPolicy:
+  """Creates a default policy from CheckpointManagerOptions."""
+  save_interval_policies = []
+  if options.should_save_fn is not None:
+    save_interval_policies.append(_ShouldSaveFnPolicy(options.should_save_fn))
+    save_interval_policies.append(
+        save_decision_policy_lib.PreemptionCheckpointingPolicy()
+    )
+  else:
+    if options.save_interval_steps is not None:
+      save_interval_policies.append(
+          save_decision_policy_lib.FixedIntervalPolicy(
+              options.save_interval_steps
+          )
+      )
+    if options.save_on_steps is not None:
+      save_interval_policies.append(
+          save_decision_policy_lib.SpecificStepsPolicy(options.save_on_steps)
+      )
+    save_interval_policies.append(
+        save_decision_policy_lib.PreemptionCheckpointingPolicy()
+    )
+    save_interval_policies.append(save_decision_policy_lib.InitialSavePolicy())
+  return save_decision_policy_lib.AnySavePolicy(save_interval_policies)
+
+
 # TODO(b/268051457) Clean up when no longer depended upon by internal users.
 def is_async_checkpointer(checkpointer: AbstractCheckpointer):
   return isinstance(
@@ -264,6 +314,12 @@ class CheckpointManagerOptions:
   temporary_path_class:
     Optional. The concrete `atomicity_types.TemporaryPath` class to be used by
     the underlying `Checkpointer`.
+  save_decision_policy: An object used to determine when a checkpoint should be
+    saved. If provided, overrides any other options dealing with this subject,
+    including `save_interval_steps`, `save_on_steps`, and `should_save_fn`, and
+    is the sole means of determining when a checkpoint should be saved. If not
+    provided, these other options are used instead. Prefer to use this option
+    over others.
   """
 
   save_interval_steps: int = 1
@@ -293,12 +349,15 @@ class CheckpointManagerOptions:
   file_options: FileOptions = dataclasses.field(default_factory=FileOptions)
   save_root_metadata: bool = True
   temporary_path_class: Optional[Type[atomicity_types.TemporaryPath]] = None
+  save_decision_policy: Optional[
+      save_decision_policy_lib.SaveDecisionPolicy
+  ] = None
 
   def __post_init__(self):
     if self.best_mode not in ('min', 'max'):
       msg = (
           "`CheckpointManagerOptions.best_mode` must be one of None, 'min' "
-          "or 'max'. Got {self.dtype}."
+          f"or 'max'. Got {self.best_mode}."
       )
       raise ValueError(msg)
     if self.max_to_keep is not None and self.max_to_keep < 0:
@@ -365,33 +424,6 @@ class CheckpointManagerOptions:
     self.save_on_steps = frozenset(self.save_on_steps or ())
 
 
-@dataclasses.dataclass
-class CheckpointInfo:
-  """Metadata about a checkpoint."""
-
-  step: int
-  time: datetime.datetime
-  metrics: Optional[PyTree]
-  is_locked: Optional[bool] = None
-
-  def __post_init__(self):
-    # Users may provide step as a jax.Array.
-    if isinstance(self.step, jax.Array):
-      self.step = int(self.step)
-
-  def __str__(self) -> str:
-    return (
-        f'Checkpoint[step={self.step} | time={self.time} |'
-        f' is_locked={self.is_locked}]'
-    )
-
-  def __eq__(self, other: CheckpointInfo) -> bool:
-    return self.step == other.step and self.time == other.time
-
-  def __hash__(self) -> int:
-    return hash((self.step, self.time))
-
-
 def _get_args_for_key(
     handler: CheckpointHandler, item_name: str
 ) -> Tuple[Type[CheckpointArgs], Type[CheckpointArgs]]:
@@ -576,6 +608,10 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
     self._options = options or CheckpointManagerOptions()
     self._multiprocessing_options = self._options.multiprocessing_options
+    self._save_decision_policy = (
+        self._options.save_decision_policy
+        or _get_default_save_decision_policy(self._options)
+    )
 
     if self._options.best_mode not in ['min', 'max']:
       raise ValueError('`best_mode` must be one of: "min", "max"')
@@ -618,12 +654,12 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       )
 
     # For async_checkpointer.
-    self._non_blocking_metadata_store = (
-        checkpoint.metadata_store(enable_write=True)
+    self._non_blocking_metadata_store = checkpoint.metadata_store(
+        enable_write=True
     )
     # For metadata checkpointer and regular checkpointer.
-    self._blocking_metadata_store = (
-        checkpoint.metadata_store(enable_write=True, blocking_write=True)
+    self._blocking_metadata_store = checkpoint.metadata_store(
+        enable_write=True, blocking_write=True
     )
 
     if checkpointers is not None:
@@ -705,7 +741,9 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         )
     )
 
-    self._checkpoints = self._load_checkpoint_infos()
+    self._checkpoints = checkpoint_info.CheckpointInfos(
+        self._load_checkpoint_infos()
+    )
 
     self._metadata_dir = self.directory / METADATA_ITEM_NAME
     if self._options.read_only and not self._metadata_dir.exists():
@@ -720,6 +758,10 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     with self._finalize_thread_lock:
       self._finalize_thread = None
 
+    self._is_saving_in_progress_lock = threading.Lock()
+    with self._is_saving_in_progress_lock:
+      self._is_saving_in_progress = False
+
     self._checkpoint_deleter: deleter.CheckpointDeleter = (
         deleter.create_checkpoint_deleter(
             self._multiprocessing_options.primary_host,
@@ -857,6 +899,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
             composite_options=composite_checkpoint_handler.CompositeOptions(
                 multiprocessing_options=options.multiprocessing_options,
                 file_options=options.file_options,
+                async_options=options.async_options,
             ),
             **item_handlers,
         ),
@@ -912,6 +955,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
             composite_options=composite_checkpoint_handler.CompositeOptions(
                 multiprocessing_options=options.multiprocessing_options,
                 file_options=options.file_options,
+                async_options=options.async_options,
             ),
             **all_item_handlers,
         ),
@@ -939,6 +983,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         CompositeCheckpointHandler(
             composite_options=composite_checkpoint_handler.CompositeOptions(
                 multiprocessing_options=options.multiprocessing_options,
+                async_options=options.async_options,
             ),
             handler_registry=handler_registry,
         ),
@@ -957,18 +1002,19 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       logging.warning(
           '`read` option is deprecated. Use `reload` to read from disk.'
       )
-      self._checkpoints = self._load_checkpoint_infos()
+      self._checkpoints.set(self._load_checkpoint_infos())
     return [ckpt.step for ckpt in self._checkpoints]
 
   def latest_step(self) -> Optional[int]:
     """See superclass documentation."""
-    return self._checkpoints[-1].step if self._checkpoints else None
+    latest = self._checkpoints.latest()
+    return latest.step if latest else None
 
   def best_step(self) -> Optional[int]:
     """See superclass documentation."""
     if not self._track_best:
       return self.latest_step()
-    if not self._checkpoints:
+    if self._checkpoints.empty():
       return None
     _, sorted_checkpoints = self._sort_checkpoints_by_metrics(self._checkpoints)
     if not sorted_checkpoints:
@@ -981,7 +1027,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     Resets internal cache of checkpoint steps, in case the directory managed
     by this object has been updated externally.
     """
-    self._checkpoints = self._load_checkpoint_infos()
+    self._checkpoints.set(self._load_checkpoint_infos())
 
   def reached_preemption(self, step: int) -> bool:
     """See superclass documentation."""
@@ -992,32 +1038,25 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     if self._options.read_only:
       logging.warning('%s is read only, save will be skipped', self.directory)
       return False
-    if self.reached_preemption(step):
-      return True
     last_checkpoint_step = self.latest_step()
     # Ensure current step is between the last step and next step (accounting for
-    # save interval). The `last_checkpoint_step` may not be initialized, in
-    # which case we should save. Otherwise, step must fall on the specified
-    # save interval. This condition accounts for the possibility of saving
-    # on preemption, in which case we want to maintain the same save period as
-    # if preemption had not happened.
+    # save interval).
     if last_checkpoint_step is not None and last_checkpoint_step >= step:
       return False
-    # If present then prefer should_save_fn over other 'save_*' options.
-    if self._options.should_save_fn is not None:
-      logging.log_every_n(
-          logging.INFO,
-          'CheckpointManagerOptions.should_save_fn is available, following save'
-          ' options will be ignored: save_interval_steps=%s and'
-          ' save_on_steps=%s',
-          500,
-          self._options.save_interval_steps,
-          self._options.save_on_steps,
-      )
-      return self._options.should_save_fn(step, last_checkpoint_step)
-    return last_checkpoint_step is None or (
-        step % self._options.save_interval_steps == 0
-        or step in self._options.save_on_steps
+
+    is_saving_in_progress = self.is_saving_in_progress()
+    reached_preemption = self.reached_preemption(step)
+    previous_step_infos = [
+        save_decision_policy_lib.StepInfo(step=ckpt.step)
+        for ckpt in self._checkpoints
+    ]
+    current_step_info = save_decision_policy_lib.StepInfo(step=step)
+    context = save_decision_policy_lib.DecisionContext(
+        is_saving_in_progress=is_saving_in_progress,
+        reached_preemption=reached_preemption,
+    )
+    return self._save_decision_policy.should_save(
+        current_step_info, previous_steps=previous_step_infos, context=context
     )
 
   def _get_save_directory(
@@ -1066,9 +1105,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         timeout=multihost.DIRECTORY_DELETION_TIMEOUT,
         processes=self._multiprocessing_options.active_processes,
     )
-    for i, info in enumerate(self._checkpoints):
-      if info.step == step:
-        self._checkpoints.pop(i)
+    self._checkpoints.delete_if(lambda info: info.step == step)
 
   def _validate_args(
       self,
@@ -1107,6 +1144,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       metrics: Optional[PyTree] = None,
       force: Optional[bool] = False,
       args: Optional[args_lib.CheckpointArgs] = None,
+      custom_metadata: dict[str, Any] | None = None,
   ) -> bool:
     """See superclass documentation."""
     process_index = multihost.process_index()
@@ -1230,7 +1268,9 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         '[process=%s] Saving checkpoint at step %d', process_index, step
     )
     step_stats.checkpointer_blocking_start_time = time.time()
-    self._checkpointer.save(save_directory, args=args)
+    self._checkpointer.save(
+        save_directory, args=args, custom_metadata=custom_metadata
+    )
     step_stats.checkpointer_blocking_duration_secs = (
         time.time() - step_stats.checkpointer_blocking_start_time
     )
@@ -1247,9 +1287,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         step_stats.get_old_steps_duration_secs,
     )
 
-    self._checkpoints = [
-        info for info in self._checkpoints if info.step not in steps_to_remove
-    ]
+    self._checkpoints.delete_if(lambda info: info.step in steps_to_remove)
     # Sync needed to ensure that old steps to remove are retrieved before
     # actually deleting them during finalize, since retrieval can involve
     # looking at the directory.
@@ -1264,6 +1302,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
     assert self._finalize_thread is None
     if is_async_checkpointer(self._checkpointer):
+      with self._is_saving_in_progress_lock:
+        self._is_saving_in_progress = True
       with self._finalize_thread_lock:
         finalize_thread_name = 'save_finalize'
         logging.info(
@@ -1398,6 +1438,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     read_step_directory = self._get_read_step_directory(step, self.directory)
 
     result = self._checkpointer.metadata(read_step_directory)
+    if isinstance(result, checkpoint.StepMetadata):
+      result = result.item_metadata
     if self._default_item is None:
       self._default_item = _determine_default_item_mode_from_directory(
           read_step_directory
@@ -1472,17 +1514,16 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       return checkpoint_infos
 
   def _get_interval_preserved_checkpoints(
-      self, checkpoints: List[CheckpointInfo]
+      self, checkpoints: checkpoint_info.CheckpointInfos
   ) -> List[CheckpointInfo]:
     """Gets which checkpoints should be kept based on keep_time_interval."""
-    if not checkpoints:
+    if checkpoints.empty():
       return []
     interval_preserved_checkpoints = [checkpoints[0]]
     if self._options.keep_time_interval is not None:
       for info in checkpoints[1:]:
-        if (
-            info.time
-            >= interval_preserved_checkpoints[-1].time
+        if info.time >= (
+            interval_preserved_checkpoints[-1].time
             + self._options.keep_time_interval
         ):
           interval_preserved_checkpoints.append(info)
@@ -1498,25 +1539,24 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
   def _metadata_file_path(self, legacy: bool = False) -> epath.Path:
     if not self._metadata_dir.exists():
       raise ValueError('Metadata directory is not initialized.')
-    return checkpoint.root_metadata_file_path(
-        self._metadata_dir, legacy=legacy
-    )
+    return checkpoint.root_metadata_file_path(self._metadata_dir, legacy=legacy)
 
   def _maybe_save_metadata(self, metadata: Mapping[str, Any]):
     """Saves CheckpointManager level metadata, skips if already present."""
     if self._options.save_root_metadata:
       logging.info('Saving root metadata')
-      if (metadata is not None and
-          not self._options.read_only and
-          utils.is_primary_host(self._multiprocessing_options.primary_host)):
+      if (
+          metadata is not None
+          and not self._options.read_only
+          and utils.is_primary_host(self._multiprocessing_options.primary_host)
+      ):
         logging.info('Creating metadata directory')
         self._metadata_dir.mkdir(parents=True, exist_ok=True)
         file_path = self._metadata_file_path()
         if not file_path.exists():  # May have been created by a previous run.
           logging.info('Writing root metadata')
           metadata_to_save = checkpoint.RootMetadata(
-              format=None,
-              custom=dict(metadata),
+              custom_metadata=dict(metadata),
           )
           self._blocking_metadata_store.write(
               file_path, serialize_root_metadata(metadata_to_save)
@@ -1536,29 +1576,31 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       if self._metadata_dir.exists():
         file_path = self._metadata_file_path()
         if not file_path.exists():
-          logging.warning('New metadata file not found in directory: %s. '
-                          'Will try to read legacy metadata file.',
-                          self._metadata_dir)
+          logging.warning(
+              'New metadata file not found in directory: %s. '
+              'Will try to read legacy metadata file.',
+              self._metadata_dir,
+          )
           file_path = self._metadata_file_path(legacy=True)
         serialized_metadata = self._blocking_metadata_store.read(file_path)
-        self._metadata = deserialize_root_metadata(serialized_metadata).custom
+        self._metadata = deserialize_root_metadata(
+            serialized_metadata
+        ).custom_metadata
         if self._metadata is None:
-          raise FileNotFoundError(
-              f'Failed to read metadata from {file_path}.'
-          )
+          raise FileNotFoundError(f'Failed to read metadata from {file_path}.')
       else:
         self._metadata = {}
     return self._metadata
 
   def _sort_checkpoints_by_metrics(
-      self, checkpoints: List[CheckpointInfo]
+      self, checkpoints: checkpoint_info.CheckpointInfos
   ) -> Tuple[List[CheckpointInfo], List[CheckpointInfo]]:
     """Sorts `checkpoints` in order of increasing metric quality.
 
     Checkpoints without corresponding metrics set will be at the beginning.
 
     Args:
-      checkpoints: a list of CheckpointInfo.
+      checkpoints: CheckpointInfos to sort.
 
     Returns:
       Tuple of CheckpointInfo lists:
@@ -1587,14 +1629,14 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     if self._options.max_to_keep is None:
       return []
     # Not enough checkpoints accumulated to consider deletion.
-    if len(self._checkpoints) <= self._options.max_to_keep:
+    if self._checkpoints.size() <= self._options.max_to_keep:
       return []
 
     # This isn't a duration but there isn't a general counter that we can use so
     # we abuse a duration metric to count the number of steps examined.
     jax.monitoring.record_event_duration_secs(
         '/jax/checkpoint/write/old_steps_examined_count',
-        len(self._checkpoints),
+        self._checkpoints.size(),
     )
 
     if self._track_best:
@@ -1606,7 +1648,7 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
     else:
       # checkpoints already sorted by ascending step
       checkpoints_without_metrics = []
-      sorted_checkpoints = self._checkpoints
+      sorted_checkpoints = [info for info in self._checkpoints]
 
     keep = int(self._options.max_to_keep)
     if self._options.keep_checkpoints_without_metrics:
@@ -1743,10 +1785,10 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         if threading.current_thread() is self._finalize_thread.save_thread():
           # If an exception occurred in the finalization of the previous
           # save, we clean up since that checkpoint was never actually saved.
-          # TODO: b/359321026 - Make self._checkpoints access threadsafe.
-          assert self._checkpoints
-          assert self._checkpoints[-1].step == step
-          self._checkpoints = self._checkpoints[:-1]
+          latest = self._checkpoints.latest()
+          assert latest is not None
+          assert latest.step == step
+          self._checkpoints.delete_if(lambda info: info.step == step)
         raise
       finally:
         # Only thread which requested save is allowed to reset Save Finalize
@@ -1766,10 +1808,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
 
   def is_saving_in_progress(self) -> bool:
     """Returns whether a checkpoint save is in progress."""
-    with self._finalize_thread_lock:
-      return (
-          self._finalize_thread is not None and self._finalize_thread.is_alive()
-      )
+    with self._is_saving_in_progress_lock:
+      return self._is_saving_in_progress
 
   def check_for_errors(self):
     """See superclass documentation."""
@@ -1802,11 +1842,11 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
       # If another checkpoint has not been previously saved, measures the time
       # since program start.
       if self.reached_preemption(step):
-        if len(self._checkpoints) > 1:
+        if self._checkpoints.size() > 1:
           previous_time = self._checkpoints[-2].time
         else:
           previous_time = _INIT_TIME
-        assert self._checkpoints
+        assert not self._checkpoints.empty()
         duration = self._checkpoints[-1].time - previous_time
         jax.monitoring.record_event_duration_secs(
             '/jax/checkpoint/write/preempt/duration_saved_secs',
@@ -1849,6 +1889,8 @@ class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
         threading.current_thread().name,
         step,
     )
+    with self._is_saving_in_progress_lock:
+      self._is_saving_in_progress = False
 
   def close(self):
     """See superclass documentation."""
